"""
prep_sampling_fft
=================
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class prep_sampling_fft(Operator):
    """Prepare time sampling optimum for FFT computation and expected
    frequencies in output.

    Parameters
    ----------
    time_freq_support : TimeFreqSupport
        Initial time domain timefreqsupport.
    cutoff_frequency : float, optional
        Cutoff frequency. in this case, number of
        points is calculated computing
        (time_range * cutoff_freq * 2) and
        taking the next power of 2 (optimum
        for fft calculation).
    number_sampling_point : int, optional
        For number of sampling point (calculation
        with cutoff_frequency is ignored).


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.mapping.prep_sampling_fft()

    >>> # Make input connections
    >>> my_time_freq_support = dpf.TimeFreqSupport()
    >>> op.inputs.time_freq_support.connect(my_time_freq_support)
    >>> my_cutoff_frequency = float()
    >>> op.inputs.cutoff_frequency.connect(my_cutoff_frequency)
    >>> my_number_sampling_point = int()
    >>> op.inputs.number_sampling_point.connect(my_number_sampling_point)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.mapping.prep_sampling_fft(
    ...     time_freq_support=my_time_freq_support,
    ...     cutoff_frequency=my_cutoff_frequency,
    ...     number_sampling_point=my_number_sampling_point,
    ... )

    >>> # Get output data
    >>> result_time_tfs_sampled = op.outputs.time_tfs_sampled()
    >>> result_freq_tfs_fft = op.outputs.freq_tfs_fft()
    """

    def __init__(
        self,
        time_freq_support=None,
        cutoff_frequency=None,
        number_sampling_point=None,
        config=None,
        server=None,
    ):
        super().__init__(name="prepare_sampling_fft", config=config, server=server)
        self._inputs = InputsPrepSamplingFft(self)
        self._outputs = OutputsPrepSamplingFft(self)
        if time_freq_support is not None:
            self.inputs.time_freq_support.connect(time_freq_support)
        if cutoff_frequency is not None:
            self.inputs.cutoff_frequency.connect(cutoff_frequency)
        if number_sampling_point is not None:
            self.inputs.number_sampling_point.connect(number_sampling_point)

    @staticmethod
    def _spec():
        description = """Prepare time sampling optimum for FFT computation and expected
            frequencies in output."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="time_freq_support",
                    type_names=["time_freq_support"],
                    optional=False,
                    document="""Initial time domain timefreqsupport.""",
                ),
                1: PinSpecification(
                    name="cutoff_frequency",
                    type_names=["double"],
                    optional=True,
                    document="""Cutoff frequency. in this case, number of
        points is calculated computing
        (time_range * cutoff_freq * 2) and
        taking the next power of 2 (optimum
        for fft calculation).""",
                ),
                2: PinSpecification(
                    name="number_sampling_point",
                    type_names=["int32"],
                    optional=True,
                    document="""For number of sampling point (calculation
        with cutoff_frequency is ignored).""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="time_tfs_sampled",
                    type_names=["time_freq_support"],
                    optional=False,
                    document="""Optimum sampled time domain timefreqsupport.""",
                ),
                1: PinSpecification(
                    name="freq_tfs_fft",
                    type_names=["time_freq_support"],
                    optional=False,
                    document="""Frequency domain timefreqsupport expected in
        output of fft.""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="prepare_sampling_fft", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsPrepSamplingFft
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsPrepSamplingFft
        """
        return super().outputs


class InputsPrepSamplingFft(_Inputs):
    """Intermediate class used to connect user inputs to
    prep_sampling_fft operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.mapping.prep_sampling_fft()
    >>> my_time_freq_support = dpf.TimeFreqSupport()
    >>> op.inputs.time_freq_support.connect(my_time_freq_support)
    >>> my_cutoff_frequency = float()
    >>> op.inputs.cutoff_frequency.connect(my_cutoff_frequency)
    >>> my_number_sampling_point = int()
    >>> op.inputs.number_sampling_point.connect(my_number_sampling_point)
    """

    def __init__(self, op: Operator):
        super().__init__(prep_sampling_fft._spec().inputs, op)
        self._time_freq_support = Input(
            prep_sampling_fft._spec().input_pin(0), 0, op, -1
        )
        self._inputs.append(self._time_freq_support)
        self._cutoff_frequency = Input(
            prep_sampling_fft._spec().input_pin(1), 1, op, -1
        )
        self._inputs.append(self._cutoff_frequency)
        self._number_sampling_point = Input(
            prep_sampling_fft._spec().input_pin(2), 2, op, -1
        )
        self._inputs.append(self._number_sampling_point)

    @property
    def time_freq_support(self):
        """Allows to connect time_freq_support input to the operator.

        Initial time domain timefreqsupport.

        Parameters
        ----------
        my_time_freq_support : TimeFreqSupport

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.mapping.prep_sampling_fft()
        >>> op.inputs.time_freq_support.connect(my_time_freq_support)
        >>> # or
        >>> op.inputs.time_freq_support(my_time_freq_support)
        """
        return self._time_freq_support

    @property
    def cutoff_frequency(self):
        """Allows to connect cutoff_frequency input to the operator.

        Cutoff frequency. in this case, number of
        points is calculated computing
        (time_range * cutoff_freq * 2) and
        taking the next power of 2 (optimum
        for fft calculation).

        Parameters
        ----------
        my_cutoff_frequency : float

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.mapping.prep_sampling_fft()
        >>> op.inputs.cutoff_frequency.connect(my_cutoff_frequency)
        >>> # or
        >>> op.inputs.cutoff_frequency(my_cutoff_frequency)
        """
        return self._cutoff_frequency

    @property
    def number_sampling_point(self):
        """Allows to connect number_sampling_point input to the operator.

        For number of sampling point (calculation
        with cutoff_frequency is ignored).

        Parameters
        ----------
        my_number_sampling_point : int

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.mapping.prep_sampling_fft()
        >>> op.inputs.number_sampling_point.connect(my_number_sampling_point)
        >>> # or
        >>> op.inputs.number_sampling_point(my_number_sampling_point)
        """
        return self._number_sampling_point


class OutputsPrepSamplingFft(_Outputs):
    """Intermediate class used to get outputs from
    prep_sampling_fft operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.mapping.prep_sampling_fft()
    >>> # Connect inputs : op.inputs. ...
    >>> result_time_tfs_sampled = op.outputs.time_tfs_sampled()
    >>> result_freq_tfs_fft = op.outputs.freq_tfs_fft()
    """

    def __init__(self, op: Operator):
        super().__init__(prep_sampling_fft._spec().outputs, op)
        self._time_tfs_sampled = Output(prep_sampling_fft._spec().output_pin(0), 0, op)
        self._outputs.append(self._time_tfs_sampled)
        self._freq_tfs_fft = Output(prep_sampling_fft._spec().output_pin(1), 1, op)
        self._outputs.append(self._freq_tfs_fft)

    @property
    def time_tfs_sampled(self):
        """Allows to get time_tfs_sampled output of the operator

        Returns
        ----------
        my_time_tfs_sampled : TimeFreqSupport

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.mapping.prep_sampling_fft()
        >>> # Connect inputs : op.inputs. ...
        >>> result_time_tfs_sampled = op.outputs.time_tfs_sampled()
        """  # noqa: E501
        return self._time_tfs_sampled

    @property
    def freq_tfs_fft(self):
        """Allows to get freq_tfs_fft output of the operator

        Returns
        ----------
        my_freq_tfs_fft : TimeFreqSupport

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.mapping.prep_sampling_fft()
        >>> # Connect inputs : op.inputs. ...
        >>> result_freq_tfs_fft = op.outputs.freq_tfs_fft()
        """  # noqa: E501
        return self._freq_tfs_fft



"""
Server factory, server configuration and communication protocols
================================================================

Contains the server factory as well as the communication
protocols and server configurations available.
"""
import logging
import os
import subprocess
import time
import io

from ansys.dpf.gate.load_api import (
    _get_path_in_install,
    _find_outdated_ansys_version,
)


class CommunicationProtocols:
    """Defines available communication protocols

    Attributes
    ----------
    gRPC = "gRPC"
        Client/Server communication via gRPC.

    InProcess = "InProcess"
        Load the DPF libraries in the Python process, communicates via a CLayer (shared memory).
    """

    gRPC = "gRPC"
    InProcess = "InProcess"


DEFAULT_COMMUNICATION_PROTOCOL = CommunicationProtocols.InProcess
DEFAULT_LEGACY = False


class DockerConfig:
    """Intermediate class encapsulating all the configuration options needed to run a docker
    image of DPF and holding tools to communicate with Docker.

    Parameters
    ----------
    use_docker : bool, optional
        Whether the DPF server should be started in a Docker Container by default.
    docker_name : str, optional
        Name of Docker Image to run.
    mounted_volumes : dict, optional
        Dictionary of key = local path and value = path of mounted volumes in the Docker Image.
        To prevent from uploading result files on the Docker Image
        :func:`ansys.dpc.core.server_factory.RunningDockerConfig.replace_with_mounted_volumes`
        iterates through this dictionary to replace local path instances by their mapped value.
    extra_args : str, optional
        Extra arguments to add to the docker run command.

    """

    def __init__(
        self,
        use_docker: bool = False,
        docker_name: str = "",
        mounted_volumes: dict = None,
        extra_args: str = "",
    ):
        from ansys.dpf.core import LOCAL_DOWNLOADED_EXAMPLES_PATH

        if mounted_volumes is None:
            mounted_volumes = {LOCAL_DOWNLOADED_EXAMPLES_PATH: "/tmp/downloaded_examples"}

        self._use_docker = use_docker
        self._docker_name = docker_name
        self._mounted_volumes = mounted_volumes
        self._extra_args = extra_args

    @property
    def use_docker(self) -> bool:
        """Whether the DPF server should be started in a Docker Container by default.

        Returns
        -------
        bool
        """
        return self._use_docker

    @use_docker.setter
    def use_docker(self, val: bool):
        self._use_docker = val

    @property
    def docker_name(self) -> str:
        """Name of Docker Image to run.

        Returns
        -------
        str
        """
        return self._docker_name

    @property
    def mounted_volumes(self) -> dict:
        """Dictionary of key = local path and value = path of mounted volumes in the Docker Image.
        To prevent from uploading result files on the Docker Image
        :func:`ansys.dpc.core.server_factory.RunningDockerConfig.replace_with_mounted_volumes`
        iterates through this dictionary to replace local path instances by their mapped value.

        Returns
        -------
        dict
        """
        return self._mounted_volumes

    @mounted_volumes.setter
    def mounted_volumes(self, mounted_volumes: dict):
        self._mounted_volumes = mounted_volumes

    @property
    def licensing_args(self) -> str:
        la = os.environ.get("ANSYS_DPF_ACCEPT_LA", "N")
        lf = os.environ.get("ANSYSLMD_LICENSE_FILE", None)
        additional_option = " -e ANSYS_DPF_ACCEPT_LA=" + la + " "
        if lf is not None:
            additional_option += " -e ANSYSLMD_LICENSE_FILE="
            additional_option += lf
            additional_option += " "
        return additional_option

    @property
    def extra_args(self) -> str:
        """Extra arguments to add to the docker run command.

        Returns
        -------
        str
        """
        return self._extra_args

    def docker_run_cmd_command(self, docker_server_port: int, local_port: int) -> str:
        """Creates the docker run command with the ``DockerConfig`` attributes as well
        as the ``docker_server_port`` and ``local_port`` passed in as parameters.

        Parameters
        ----------
        docker_server_port : int
            Port used inside the Docker Container to run the gRPC server.
        local_port : int
            Port exposed outside the Docker container bounded to the internal
            ``docker_server_port``.

        Returns
        -------
        str
        """
        mounted_volumes_args = "-v " + " -v ".join(
            key + ":" + val for key, val in self.mounted_volumes.items()
        )
        licensing_options = self.licensing_args
        return (
            f"docker run -d "
            f" {licensing_options} "
            f"-p {local_port}:{docker_server_port} "
            f"{self.extra_args} "
            f"{mounted_volumes_args} "
            f"-e DOCKER_SERVER_PORT={docker_server_port} "
            f"--expose={docker_server_port} "
            f"{self.docker_name}"
        )

    def __str__(self):
        return (
            "DockerConfig with: \n"
            f"\t- use_docker: {self.use_docker}\n"
            f"\t- docker_name: {self.docker_name}\n"
            f"\t- mounted_volume: {self.mounted_volumes}\n"
            f"\t- extra_args: {self.extra_args}\n"
        )

    @staticmethod
    def find_port_available_for_docker_bind(port: int) -> int:
        """Checks for available internal ``docker_server_port`` by looking at the stdout of
        all running Docker Containers.

        Parameters
        ----------
        port: int

        Returns
        -------
        port: int
        """
        run_cmd = "docker ps --all"
        b_shell = False
        if os.name == "posix":
            b_shell = True
        with subprocess.Popen(
            run_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=b_shell
        ) as process:
            used_ports = []
            with io.TextIOWrapper(process.stdout, encoding="utf-8") as log_out:
                for line in log_out:
                    if not ("CONTAINER ID" in line):
                        split = line.split("0.0.0.0:")
                        if len(split) > 1:
                            used_ports.append(int(split[1].split("-")[0]))

            while port in used_ports:
                port += 1
        return port


class ServerConfig:
    """Provides an instance of ServerConfig object to manage the server type used.
    The default parameters can be overwritten using the DPF_SERVER_TYPE environment
    variable. DPF_SERVER_TYPE=INPROCESS, DPF_SERVER_TYPE=GRPC,
    DPF_SERVER_TYPE=LEGACYGRPC can be used.

    Parameters
    ----------
    protocol : CommunicationProtocols, optional
        Communication protocol for DPF server (e.g. InProcess, gRPC)
    legacy : bool, optional
        If legacy is set to True, the server will be using ansys-grpc-dpf
        Python module. If not, it will communicate with DPF binaries using ctypes
        and DPF CLayer calls.

    Examples
    --------
    Use constructor parameters to manually create servers.

    >>> from ansys.dpf import core as dpf
    >>> in_process_config = dpf.ServerConfig(
    ...     protocol=dpf.server_factory.CommunicationProtocols.InProcess, legacy=False)
    >>> grpc_config = dpf.ServerConfig(
    ...     protocol=dpf.server_factory.CommunicationProtocols.gRPC, legacy=False)
    >>> legacy_grpc_config = dpf.ServerConfig(
    ...     protocol=dpf.server_factory.CommunicationProtocols.gRPC, legacy=True)
    >>> in_process_server = dpf.start_local_server(config=in_process_config, as_global=False)
    >>> grpc_server = dpf.start_local_server(config=grpc_config, as_global=False)
    >>> legacy_grpc_server = dpf.start_local_server(config=legacy_grpc_config, as_global=False)

    Use the environment variable to set the default server configuration.

    >>> import os
    >>> os.environ["DPF_SERVER_TYPE"] = "INPROCESS"
    >>> dpf.start_local_server()
    <ansys.dpf.core.server_types.InProcessServer object at ...>

    """

    def __init__(
        self,
        protocol: str = DEFAULT_COMMUNICATION_PROTOCOL,
        legacy: bool = DEFAULT_LEGACY,
    ):
        self.legacy = legacy
        if not protocol:
            self.protocol = CommunicationProtocols.InProcess
        else:
            self.protocol = protocol

    def __str__(self):
        text = f"Server configuration: protocol={self.protocol}"
        if self.legacy:
            text += f" (legacy gRPC)"
        return text

    def __eq__(self, other: "ServerConfig"):
        if isinstance(other, ServerConfig):
            return self.legacy == other.legacy and self.protocol == other.protocol
        return False

    def __ne__(self, other):
        return not self.__eq__(other)


def get_default_server_config(
    server_lower_than_or_equal_to_0_3: bool = False, docker_config: DockerConfig = None
):
    """Returns the default configuration depending on the server version.
    - if ansys.dpf.core.SERVER_CONFIGURATION is not None, then this variable is taken
    - if server_lower_than_or_equal_to_0_3 is True, then LegacyGrpcServer is taken
    - DPF_SERVER_TYPE environment variable is set to ``INPROCESS``, ``GRPC``, ``LEGACYGRPC``,
      then this variable is taken
    else DEFAULT_COMMUNICATION_PROTOCOL and DEFAULT_LEGACY is taken.

    Raises
    ------
    If DPF_SERVER_TYPE environment variable is set to unknown value.

    """
    from ansys.dpf.core import SERVER_CONFIGURATION

    if SERVER_CONFIGURATION is not None:
        return SERVER_CONFIGURATION

    if docker_config is None:
        docker_config = DockerConfig()

    config = None

    if server_lower_than_or_equal_to_0_3:
        return ServerConfig(protocol=CommunicationProtocols.gRPC, legacy=True)

    dpf_server_type = os.environ.get("DPF_SERVER_TYPE", None)
    if dpf_server_type and config is None:
        if dpf_server_type == "INPROCESS":
            config = AvailableServerConfigs.InProcessServer
        elif dpf_server_type == "GRPC":
            config = AvailableServerConfigs.GrpcServer
        elif dpf_server_type == "LEGACYGRPC":
            config = AvailableServerConfigs.LegacyGrpcServer
        else:
            raise NotImplementedError(
                f"DPF_SERVER_TYPE environment variable must "
                f"be set to one of the following: INPROCESS, "
                f"GRPC, LEGACYGRPC."
            )
    elif config is None and docker_config.use_docker:
        config = get_default_remote_server_config()
    elif config is None:
        config = ServerConfig(protocol=DEFAULT_COMMUNICATION_PROTOCOL, legacy=DEFAULT_LEGACY)
    return config


def get_default_remote_server_config():
    """Returns the default configuration for gRPC communication.
    Follows get_default_server_config

    Raises
    ------
    If DPF_SERVER_TYPE environment variable is set to unknown value.

    """
    config = get_default_server_config()
    if config == AvailableServerConfigs.InProcessServer:
        return AvailableServerConfigs.GrpcServer


class AvailableServerConfigs:
    """Defines available server configurations

    Attributes
    ----------
    LegacyGrpcServer = ServerConfig(CommunicationProtocols.gRPC, legacy=True)
       Using gRPC communication through the python module ansys.grpc.dpf.

    InProcess = ServerConfig(CommunicationProtocols.InProcess, legacy=False)
        Loading DPF in Process.

    GrpcServer = ServerConfig(CommunicationProtocols.gRPC, legacy=False)
        Using gRPC communication through DPF gRPC CLayer Ans.Dpf.GrpcClient.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> in_process_config = dpf.AvailableServerConfigs.InProcessServer
    >>> grpc_config = dpf.AvailableServerConfigs.GrpcServer
    >>> legacy_grpc_config = dpf.AvailableServerConfigs.LegacyGrpcServer
    >>> in_process_server = dpf.start_local_server(config=in_process_config, as_global=False)
    >>> grpc_server = dpf.start_local_server(config=grpc_config, as_global=False)
    >>> legacy_grpc_server = dpf.start_local_server(config=legacy_grpc_config, as_global=False)

    """

    LegacyGrpcServer = ServerConfig(CommunicationProtocols.gRPC, legacy=True)
    InProcessServer = ServerConfig(CommunicationProtocols.InProcess, legacy=False)
    GrpcServer = ServerConfig(CommunicationProtocols.gRPC, legacy=False)


class RunningDockerConfig:
    """Holds all the configuration options and the process information of a running Docker image
    of a DPF server.

    Parameters
    ----------
    docker_config : DockerConfig, optional
        ``DockerConfig`` used to start the docker.
    server_id : int, optional
        Running Docker Container id.
    docker_server_port : int, optional
        Local port exposed to the docker image.

    """

    def __init__(
        self,
        docker_config: DockerConfig = None,
        server_id: int = None,
        docker_server_port: int = None,
    ):
        if docker_config is None:
            docker_config = DockerConfig()
        self._docker_config = docker_config
        self._server_id = server_id
        self._use_docker = self._docker_config.use_docker
        self._port = docker_server_port

    @property
    def use_docker(self) -> bool:
        """Whether the DPF server should be started in a Docker Container by default.

        Returns
        -------
        bool
        """
        return self._use_docker

    @property
    def docker_server_port(self) -> int:
        """Port used inside the Docker Container to run the gRPC server.

        Returns
        -------
        int
        """
        return self._port

    @docker_server_port.setter
    def docker_server_port(self, val: int):
        self._port = val

    @property
    def server_id(self) -> int:
        """Running Docker Container id.

        Returns
        -------
        int
        """
        return self._server_id

    @server_id.setter
    def server_id(self, val: int):
        self._server_id = val

    @property
    def docker_name(self) -> str:
        """Name of Docker running Image.

        Returns
        -------
        str
        """
        return self._docker_config.docker_name

    @property
    def mounted_volumes(self) -> dict:
        """Dictionary of local path to docker path of volumes mounted in the Docker Image.
        These paths are checked for when result files are looked for by the server to prevent from
        uploading them.

        Returns
        -------
        dict
        """
        return self._docker_config.mounted_volumes

    @property
    def extra_args(self) -> str:
        """Extra arguments used in the ``docker run`` command

        Returns
        -------
        str
        """
        return self._docker_config.extra_args

    def replace_with_mounted_volumes(self, path: str) -> str:
        """Replace local path found in the list of mounted
        volumes by their mounted path in the docker.

        Parameters
        ----------
        path: str
            Path to search for occurrences of mounted volumes.

        Returns
        -------
        path: str

        """
        if self.use_docker:
            path = os.path.normpath(path)
            for key, val in self.mounted_volumes.items():
                path = path.replace(os.path.normpath(key), val)
        return path

    def remove_docker_image(self) -> None:
        """Stops and Removes the Docker image with its id==server_id"""
        if not self.use_docker or not self.server_id:
            return
        stop_cmd = f"docker stop {self.server_id}"
        b_shell = False
        if os.name == "posix":
            b_shell = True
        with subprocess.Popen(
            stop_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=b_shell
        ) as process:
            rm_cmd = f"docker rm {self.server_id}"
            with io.TextIOWrapper(process.stdout, encoding="utf-8") as log_out:
                for _ in log_out:
                    pass
            try:
                subprocess.run(
                    rm_cmd,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    shell=b_shell,
                    check=True,
                )
            except subprocess.CalledProcessError as e:
                if "No such container" in str(e.output):
                    pass
                else:
                    raise e
            process.kill()

    def listen_to_process(
        self,
        log: logging.Logger,
        cmd_lines: list,
        lines: list,
        timeout: float,
        stdout: bool = True,
    ) -> None:
        """Search inside the Docker Container stdout log to fill in this instance's attributes.

        Parameters
        ----------
        log
            Instance of ``logging`` to add debug info to.
        cmd_lines: list
            Stdout of the shell process run ``docker run`` command.
        lines : list
            Internal Container's stdout are copied into ``lines``.
        timeout : float
            When to stop searching for stdout.
        stdout : bool, optional
            Whether to check stdout or stderr.
        """
        self.server_id = cmd_lines[0].replace("\n", "")
        t_timeout = time.time() + timeout
        while time.time() < t_timeout:
            with subprocess.Popen(
                f"docker logs {self.server_id}",
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                shell=(os.name == "posix"),
            ) as docker_process:
                self._use_docker = True
                if stdout:
                    with io.TextIOWrapper(docker_process.stdout, encoding="utf-8") as log_out:
                        for line in log_out:
                            log.debug(line)
                            lines.append(line)
                else:
                    with io.TextIOWrapper(docker_process.stderr, encoding="utf-8") as log_error:
                        for line in log_error:
                            if line not in lines:
                                lines.append(line)
                docker_process.kill()

    def docker_run_cmd_command(self, docker_server_port: int, local_port: int) -> str:
        return self._docker_config.docker_run_cmd_command(docker_server_port, local_port)

    def __str__(self):
        return str(self._docker_config) + f"\t- server_id: {self.server_id}\n"


def create_default_docker_config() -> DockerConfig:
    return DockerConfig(
        use_docker="DPF_DOCKER" in os.environ.keys(),
        docker_name=os.environ.get("DPF_DOCKER", ""),
    )


class ServerFactory:
    """Factory for server type choice depending on current configuration."""

    @staticmethod
    def get_server_type_from_config(
        config: ServerConfig = None,
        ansys_path: str = None,
        docker_config: DockerConfig = None,
    ):
        from ansys.dpf.core.server_types import (
            LegacyGrpcServer,
            GrpcServer,
            InProcessServer,
        )

        # dpf.core.SERVER_CONFIGURATION is required to know what type of connection to set
        if config is None:
            # If no SERVER_CONFIGURATION is yet defined, set one with default values
            is_server_old = False
            if ansys_path is not None:
                if "ansys_dpf_server" not in ansys_path:
                    is_server_old = _find_outdated_ansys_version(ansys_path)
            config = get_default_server_config(is_server_old, docker_config)
        if config.protocol == CommunicationProtocols.gRPC and config.legacy:
            return LegacyGrpcServer
        elif config.protocol == CommunicationProtocols.gRPC and not config.legacy:
            if ansys_path is None:
                from ansys.dpf.core.misc import __ansys_version__

                ansys_path = os.environ.get("AWP_ROOT" + str(__ansys_version__), None)
            if ansys_path is not None:
                sub_folders = os.path.join(ansys_path, _get_path_in_install())
                os.environ["PATH"] += sub_folders
            return GrpcServer
        elif config.protocol == CommunicationProtocols.InProcess and not config.legacy:
            return InProcessServer
        else:
            raise NotImplementedError("Server config not available.")

    @staticmethod
    def get_remote_server_type_from_config(config: ServerConfig = None):
        if config is None:
            config = get_default_remote_server_config()
        return ServerFactory.get_server_type_from_config(config)



"""
path_utilities
==============
Offer tools similar to os.path but taking the os of the
server into account to create path.
"""

import os

import ansys.dpf.core.server_types
from ansys.dpf.core import server as server_module
from pathlib import Path


def join(*args, **kwargs):
    """Join two strings to form a path, following the server
    architecture.
    Using a server version below 3.0, please ensure that the
    python client and the server's os are similar before
    using this method.

    Parameters
    ----------
    args : str, os.PathLike, LegacyGrpcServer
        Path to join and optionally a server.

    kwargs : LegacyGrpcServer
        server=.

    server : Server
        Specific server to use.

    Returns
    -------
    concatenated_file_path : str
        left_path + right_path concatenated into a single string value.

    """
    server = None
    parts = []
    for a in args:
        if isinstance(a, (str, Path)) and len(a) > 0:
            parts.append(str(a))
        elif isinstance(a, ansys.dpf.core.server_types.LegacyGrpcServer):
            server = a
    if "server" in kwargs:
        server = kwargs["server"]
    if not server:
        server = server_module._global_server()
    if not server:
        if ansys.dpf.core.server_types.RUNNING_DOCKER.use_docker:
            current_os = "posix"
        else:
            return os.path.join(*args)
    else:
        current_os = server.os

    if len(parts) == 0:
        return ""
    separator = "\\"
    if current_os == "posix":
        separator = "/"
    path_to_return = parts[0]
    for ipath in range(1, len(parts)):
        path_to_return += separator + parts[ipath]
    return path_to_return


def to_server_os(path, server=None):
    path = str(path)
    server = server_module.get_or_create_server(server)
    path = server.docker_config.replace_with_mounted_volumes(path)
    if server.os == "posix":
        return path.replace("\\", "/")
    else:
        return path.replace("/", "\\")



"""
members_in_compression_not_certified
====================================
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class members_in_compression_not_certified(Operator):
    """This operator is a non-certified example of buckling resistance
    verification for the compression members for Class I, 2 and 3
    cross-sections. It is only provided as an example if you want to
    develop your own compute norm operator. The results computed by
    this beta operator have not been certified by ANSYS. ANSYS
    declines all responsibility for the use of this operator.

    Parameters
    ----------
    time_scoping : Scoping or int, optional
        Time/freq set ids (use ints or scoping)
    field_yield_strength : DataSources or Field
        This pin contains file csv or field of beam's
        yield strength.
    field_end_condition : DataSources or Field
        This pin contains file csv or field of beam's
        end condition defined by the user. if
        no input at this pin found, it would
        take end condition's value of all
        beams as 1.
    streams : StreamsContainer, optional
        result file container allowed to be kept open
        to cache data.
    data_sources : DataSources, optional
        Result file path container, used if no
        streams are set.
    manufacture : bool
        Manufacturing processus:hot finished if true
        or cold formed if false. default
        value : hot finished.
    partial_factor : float
        Partial safety factor for resistance of
        members to instability assessed by
        member checks. default value: 1.
    mesh : MeshedRegion
        mesh containing beam's properties defined by
        user
    axial_force : FieldsContainer
        Fields container of axial force defined by
        user
    fabrication_type : bool
        If there is beam i in the structure, please
        define its fabrication type. true:
        rolled section, false: welded section


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.result.members_in_compression_not_certified()

    >>> # Make input connections
    >>> my_time_scoping = dpf.Scoping()
    >>> op.inputs.time_scoping.connect(my_time_scoping)
    >>> my_field_yield_strength = dpf.DataSources()
    >>> op.inputs.field_yield_strength.connect(my_field_yield_strength)
    >>> my_field_end_condition = dpf.DataSources()
    >>> op.inputs.field_end_condition.connect(my_field_end_condition)
    >>> my_streams = dpf.StreamsContainer()
    >>> op.inputs.streams.connect(my_streams)
    >>> my_data_sources = dpf.DataSources()
    >>> op.inputs.data_sources.connect(my_data_sources)
    >>> my_manufacture = bool()
    >>> op.inputs.manufacture.connect(my_manufacture)
    >>> my_partial_factor = float()
    >>> op.inputs.partial_factor.connect(my_partial_factor)
    >>> my_mesh = dpf.MeshedRegion()
    >>> op.inputs.mesh.connect(my_mesh)
    >>> my_axial_force = dpf.FieldsContainer()
    >>> op.inputs.axial_force.connect(my_axial_force)
    >>> my_fabrication_type = bool()
    >>> op.inputs.fabrication_type.connect(my_fabrication_type)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.result.members_in_compression_not_certified(
    ...     time_scoping=my_time_scoping,
    ...     field_yield_strength=my_field_yield_strength,
    ...     field_end_condition=my_field_end_condition,
    ...     streams=my_streams,
    ...     data_sources=my_data_sources,
    ...     manufacture=my_manufacture,
    ...     partial_factor=my_partial_factor,
    ...     mesh=my_mesh,
    ...     axial_force=my_axial_force,
    ...     fabrication_type=my_fabrication_type,
    ... )

    >>> # Get output data
    >>> result_buckling_resistance_compression_yy = op.outputs.buckling_resistance_compression_yy()
    >>> result_buckling_resistance_compression_zz = op.outputs.buckling_resistance_compression_zz()
    """

    def __init__(
        self,
        time_scoping=None,
        field_yield_strength=None,
        field_end_condition=None,
        streams=None,
        data_sources=None,
        manufacture=None,
        partial_factor=None,
        mesh=None,
        axial_force=None,
        fabrication_type=None,
        config=None,
        server=None,
    ):
        super().__init__(
            name="members_in_compression_not_certified", config=config, server=server
        )
        self._inputs = InputsMembersInCompressionNotCertified(self)
        self._outputs = OutputsMembersInCompressionNotCertified(self)
        if time_scoping is not None:
            self.inputs.time_scoping.connect(time_scoping)
        if field_yield_strength is not None:
            self.inputs.field_yield_strength.connect(field_yield_strength)
        if field_end_condition is not None:
            self.inputs.field_end_condition.connect(field_end_condition)
        if streams is not None:
            self.inputs.streams.connect(streams)
        if data_sources is not None:
            self.inputs.data_sources.connect(data_sources)
        if manufacture is not None:
            self.inputs.manufacture.connect(manufacture)
        if partial_factor is not None:
            self.inputs.partial_factor.connect(partial_factor)
        if mesh is not None:
            self.inputs.mesh.connect(mesh)
        if axial_force is not None:
            self.inputs.axial_force.connect(axial_force)
        if fabrication_type is not None:
            self.inputs.fabrication_type.connect(fabrication_type)

    @staticmethod
    def _spec():
        description = """This operator is a non-certified example of buckling resistance
            verification for the compression members for Class I, 2
            and 3 cross-sections. It is only provided as an example if
            you want to develop your own compute norm operator. The
            results computed by this beta operator have not been
            certified by ANSYS. ANSYS declines all responsibility for
            the use of this operator."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="time_scoping",
                    type_names=["scoping", "vector<int32>", "int32"],
                    optional=True,
                    document="""Time/freq set ids (use ints or scoping)""",
                ),
                1: PinSpecification(
                    name="field_yield_strength",
                    type_names=["data_sources", "field"],
                    optional=False,
                    document="""This pin contains file csv or field of beam's
        yield strength.""",
                ),
                2: PinSpecification(
                    name="field_end_condition",
                    type_names=["data_sources", "field"],
                    optional=False,
                    document="""This pin contains file csv or field of beam's
        end condition defined by the user. if
        no input at this pin found, it would
        take end condition's value of all
        beams as 1.""",
                ),
                3: PinSpecification(
                    name="streams",
                    type_names=["streams_container"],
                    optional=True,
                    document="""result file container allowed to be kept open
        to cache data.""",
                ),
                4: PinSpecification(
                    name="data_sources",
                    type_names=["data_sources"],
                    optional=True,
                    document="""Result file path container, used if no
        streams are set.""",
                ),
                5: PinSpecification(
                    name="manufacture",
                    type_names=["bool"],
                    optional=False,
                    document="""Manufacturing processus:hot finished if true
        or cold formed if false. default
        value : hot finished.""",
                ),
                6: PinSpecification(
                    name="partial_factor",
                    type_names=["double"],
                    optional=False,
                    document="""Partial safety factor for resistance of
        members to instability assessed by
        member checks. default value: 1.""",
                ),
                7: PinSpecification(
                    name="mesh",
                    type_names=["abstract_meshed_region"],
                    optional=False,
                    document="""mesh containing beam's properties defined by
        user""",
                ),
                8: PinSpecification(
                    name="axial_force",
                    type_names=["fields_container"],
                    optional=False,
                    document="""Fields container of axial force defined by
        user""",
                ),
                12: PinSpecification(
                    name="fabrication_type",
                    type_names=["bool"],
                    optional=False,
                    document="""If there is beam i in the structure, please
        define its fabrication type. true:
        rolled section, false: welded section""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="buckling_resistance_compression_yy",
                    type_names=["fields_container"],
                    optional=False,
                    document="""Fields container of buckling resistance
        factor on axis y-y in case of
        compression. these factors should be
        less than 1 and positive.""",
                ),
                1: PinSpecification(
                    name="buckling_resistance_compression_zz",
                    type_names=["fields_container"],
                    optional=False,
                    document="""Fields container of buckling resistance
        factor on axis z-z in case of
        compression. these factors should be
        less than 1 and positive.""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(
            name="members_in_compression_not_certified", server=server
        )

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsMembersInCompressionNotCertified
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsMembersInCompressionNotCertified
        """
        return super().outputs


class InputsMembersInCompressionNotCertified(_Inputs):
    """Intermediate class used to connect user inputs to
    members_in_compression_not_certified operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.result.members_in_compression_not_certified()
    >>> my_time_scoping = dpf.Scoping()
    >>> op.inputs.time_scoping.connect(my_time_scoping)
    >>> my_field_yield_strength = dpf.DataSources()
    >>> op.inputs.field_yield_strength.connect(my_field_yield_strength)
    >>> my_field_end_condition = dpf.DataSources()
    >>> op.inputs.field_end_condition.connect(my_field_end_condition)
    >>> my_streams = dpf.StreamsContainer()
    >>> op.inputs.streams.connect(my_streams)
    >>> my_data_sources = dpf.DataSources()
    >>> op.inputs.data_sources.connect(my_data_sources)
    >>> my_manufacture = bool()
    >>> op.inputs.manufacture.connect(my_manufacture)
    >>> my_partial_factor = float()
    >>> op.inputs.partial_factor.connect(my_partial_factor)
    >>> my_mesh = dpf.MeshedRegion()
    >>> op.inputs.mesh.connect(my_mesh)
    >>> my_axial_force = dpf.FieldsContainer()
    >>> op.inputs.axial_force.connect(my_axial_force)
    >>> my_fabrication_type = bool()
    >>> op.inputs.fabrication_type.connect(my_fabrication_type)
    """

    def __init__(self, op: Operator):
        super().__init__(members_in_compression_not_certified._spec().inputs, op)
        self._time_scoping = Input(
            members_in_compression_not_certified._spec().input_pin(0), 0, op, -1
        )
        self._inputs.append(self._time_scoping)
        self._field_yield_strength = Input(
            members_in_compression_not_certified._spec().input_pin(1), 1, op, -1
        )
        self._inputs.append(self._field_yield_strength)
        self._field_end_condition = Input(
            members_in_compression_not_certified._spec().input_pin(2), 2, op, -1
        )
        self._inputs.append(self._field_end_condition)
        self._streams = Input(
            members_in_compression_not_certified._spec().input_pin(3), 3, op, -1
        )
        self._inputs.append(self._streams)
        self._data_sources = Input(
            members_in_compression_not_certified._spec().input_pin(4), 4, op, -1
        )
        self._inputs.append(self._data_sources)
        self._manufacture = Input(
            members_in_compression_not_certified._spec().input_pin(5), 5, op, -1
        )
        self._inputs.append(self._manufacture)
        self._partial_factor = Input(
            members_in_compression_not_certified._spec().input_pin(6), 6, op, -1
        )
        self._inputs.append(self._partial_factor)
        self._mesh = Input(
            members_in_compression_not_certified._spec().input_pin(7), 7, op, -1
        )
        self._inputs.append(self._mesh)
        self._axial_force = Input(
            members_in_compression_not_certified._spec().input_pin(8), 8, op, -1
        )
        self._inputs.append(self._axial_force)
        self._fabrication_type = Input(
            members_in_compression_not_certified._spec().input_pin(12), 12, op, -1
        )
        self._inputs.append(self._fabrication_type)

    @property
    def time_scoping(self):
        """Allows to connect time_scoping input to the operator.

        Time/freq set ids (use ints or scoping)

        Parameters
        ----------
        my_time_scoping : Scoping or int

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.members_in_compression_not_certified()
        >>> op.inputs.time_scoping.connect(my_time_scoping)
        >>> # or
        >>> op.inputs.time_scoping(my_time_scoping)
        """
        return self._time_scoping

    @property
    def field_yield_strength(self):
        """Allows to connect field_yield_strength input to the operator.

        This pin contains file csv or field of beam's
        yield strength.

        Parameters
        ----------
        my_field_yield_strength : DataSources or Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.members_in_compression_not_certified()
        >>> op.inputs.field_yield_strength.connect(my_field_yield_strength)
        >>> # or
        >>> op.inputs.field_yield_strength(my_field_yield_strength)
        """
        return self._field_yield_strength

    @property
    def field_end_condition(self):
        """Allows to connect field_end_condition input to the operator.

        This pin contains file csv or field of beam's
        end condition defined by the user. if
        no input at this pin found, it would
        take end condition's value of all
        beams as 1.

        Parameters
        ----------
        my_field_end_condition : DataSources or Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.members_in_compression_not_certified()
        >>> op.inputs.field_end_condition.connect(my_field_end_condition)
        >>> # or
        >>> op.inputs.field_end_condition(my_field_end_condition)
        """
        return self._field_end_condition

    @property
    def streams(self):
        """Allows to connect streams input to the operator.

        result file container allowed to be kept open
        to cache data.

        Parameters
        ----------
        my_streams : StreamsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.members_in_compression_not_certified()
        >>> op.inputs.streams.connect(my_streams)
        >>> # or
        >>> op.inputs.streams(my_streams)
        """
        return self._streams

    @property
    def data_sources(self):
        """Allows to connect data_sources input to the operator.

        Result file path container, used if no
        streams are set.

        Parameters
        ----------
        my_data_sources : DataSources

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.members_in_compression_not_certified()
        >>> op.inputs.data_sources.connect(my_data_sources)
        >>> # or
        >>> op.inputs.data_sources(my_data_sources)
        """
        return self._data_sources

    @property
    def manufacture(self):
        """Allows to connect manufacture input to the operator.

        Manufacturing processus:hot finished if true
        or cold formed if false. default
        value : hot finished.

        Parameters
        ----------
        my_manufacture : bool

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.members_in_compression_not_certified()
        >>> op.inputs.manufacture.connect(my_manufacture)
        >>> # or
        >>> op.inputs.manufacture(my_manufacture)
        """
        return self._manufacture

    @property
    def partial_factor(self):
        """Allows to connect partial_factor input to the operator.

        Partial safety factor for resistance of
        members to instability assessed by
        member checks. default value: 1.

        Parameters
        ----------
        my_partial_factor : float

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.members_in_compression_not_certified()
        >>> op.inputs.partial_factor.connect(my_partial_factor)
        >>> # or
        >>> op.inputs.partial_factor(my_partial_factor)
        """
        return self._partial_factor

    @property
    def mesh(self):
        """Allows to connect mesh input to the operator.

        mesh containing beam's properties defined by
        user

        Parameters
        ----------
        my_mesh : MeshedRegion

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.members_in_compression_not_certified()
        >>> op.inputs.mesh.connect(my_mesh)
        >>> # or
        >>> op.inputs.mesh(my_mesh)
        """
        return self._mesh

    @property
    def axial_force(self):
        """Allows to connect axial_force input to the operator.

        Fields container of axial force defined by
        user

        Parameters
        ----------
        my_axial_force : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.members_in_compression_not_certified()
        >>> op.inputs.axial_force.connect(my_axial_force)
        >>> # or
        >>> op.inputs.axial_force(my_axial_force)
        """
        return self._axial_force

    @property
    def fabrication_type(self):
        """Allows to connect fabrication_type input to the operator.

        If there is beam i in the structure, please
        define its fabrication type. true:
        rolled section, false: welded section

        Parameters
        ----------
        my_fabrication_type : bool

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.members_in_compression_not_certified()
        >>> op.inputs.fabrication_type.connect(my_fabrication_type)
        >>> # or
        >>> op.inputs.fabrication_type(my_fabrication_type)
        """
        return self._fabrication_type


class OutputsMembersInCompressionNotCertified(_Outputs):
    """Intermediate class used to get outputs from
    members_in_compression_not_certified operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.result.members_in_compression_not_certified()
    >>> # Connect inputs : op.inputs. ...
    >>> result_buckling_resistance_compression_yy = op.outputs.buckling_resistance_compression_yy()
    >>> result_buckling_resistance_compression_zz = op.outputs.buckling_resistance_compression_zz()
    """

    def __init__(self, op: Operator):
        super().__init__(members_in_compression_not_certified._spec().outputs, op)
        self._buckling_resistance_compression_yy = Output(
            members_in_compression_not_certified._spec().output_pin(0), 0, op
        )
        self._outputs.append(self._buckling_resistance_compression_yy)
        self._buckling_resistance_compression_zz = Output(
            members_in_compression_not_certified._spec().output_pin(1), 1, op
        )
        self._outputs.append(self._buckling_resistance_compression_zz)

    @property
    def buckling_resistance_compression_yy(self):
        """Allows to get buckling_resistance_compression_yy output of the operator

        Returns
        ----------
        my_buckling_resistance_compression_yy : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.members_in_compression_not_certified()
        >>> # Connect inputs : op.inputs. ...
        >>> result_buckling_resistance_compression_yy = op.outputs.buckling_resistance_compression_yy()
        """  # noqa: E501
        return self._buckling_resistance_compression_yy

    @property
    def buckling_resistance_compression_zz(self):
        """Allows to get buckling_resistance_compression_zz output of the operator

        Returns
        ----------
        my_buckling_resistance_compression_zz : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.members_in_compression_not_certified()
        >>> # Connect inputs : op.inputs. ...
        >>> result_buckling_resistance_compression_zz = op.outputs.buckling_resistance_compression_zz()
        """  # noqa: E501
        return self._buckling_resistance_compression_zz



from typing import NamedTuple


def class_handling_cache(cls):
    """Class decorator used to handle cache.
    To use it, add a ''_to_cache'' static attribute in the given class.
    This private dictionary should map class getters to their list of setters.
    At initialization, this decorator add a ''_cache'' property to the class.
    This new property is an instance of ''CacheHandler''.

    .. note::
       The method must be used as a class decorator.
    """
    if hasattr(cls, "_to_cache"):

        def get_handler(mesh):
            if hasattr(mesh, "__cache"):
                return mesh.__cache
            else:
                setattr(mesh, "__cache", CacheHandler(cls, cls._to_cache))
            return mesh.__cache

        for getter, setters in cls._to_cache.items():
            if setters:
                for setter in setters:
                    setattr(cls, setter.__name__, _handle_cache(setter))
            setattr(cls, getter.__name__, _handle_cache(getter))

        setattr(cls, "_cache", property(get_handler))
    return cls


class MethodIdentifier(NamedTuple):
    method_name: str
    args: list
    kwargs: dict

    def __eq__(self, other):
        if isinstance(other, str):
            return self.method_name == other
        else:
            return (
                self.method_name == other.method_name
                and self.args == other.args
                and self.kwargs == other.kwargs
            )

    def __hash__(self):
        hash = self.method_name.__hash__()
        if self.args:
            hash += self.args.__hash__()
        if self.kwargs:
            hash += self.kwargs.__hash__()
        return hash


class CacheHandler:
    """ "Handle cache complexity.
    Is initialized by a class and a dictionary mapping the getters
    which support caching to their setters.
    When the getters of the dictionary are called, their parameters
    and their results are cached so that, when the getters are called again
    with the same parameters, the data is directly recovered instead of reevaluated.
    When the setters associated to getters in the input dictionary are called,
    their associated getters' caches are cleared.

    Parameters
    ----------
    cls : type
        Class type

    getters_to_setters_dict : dict[function:list[function]]
        Map class getters to their list of setters which need to be cached
    """

    def __init__(self, cls, getters_to_setters_dict):

        self.getter_to_setters_name = {}
        for getter, setters in getters_to_setters_dict.items():
            setters_name = []
            if setters:
                for setter in setters:
                    setters_name.append(setter.__name__)
            self.getter_to_setters_name[getter.__name__] = setters_name

        self.setter_to_getter_names = {}
        for getter, setters in self.getter_to_setters_name.items():
            for setter in setters:
                self.setter_to_getter_names[setter] = getter

        self.cached = {}

    def handle(self, object, func, *args, **kwargs):
        identifier = MethodIdentifier(func.__name__, args, kwargs)
        if identifier in self.cached:
            return self.cached[identifier]
        elif func.__name__ in self.getter_to_setters_name:
            self.cached[identifier] = func(object, *args, **kwargs)
            setattr(func, "under_cache", False)
            return self.cached[identifier]
        else:
            if (
                func.__name__ in self.setter_to_getter_names
                and self.setter_to_getter_names[func.__name__] in self.cached
            ):
                del self.cached[self.setter_to_getter_names[func.__name__]]
            return func(object, *args, **kwargs)

    def clear(self):
        self.cached = {}


def _handle_cache(func):
    """Calls the cache handler to either recover cached data, either cache the data
    or clear some cached data if the method is a setter.

    .. note::
       The method must be used as a decorator.
    """

    def wrapper(self, *args, **kwargs):
        """Call the original function"""
        if hasattr(self, "_cache"):
            return self._cache.handle(self, func, *args, **kwargs)
        else:
            func(self, *args, **kwargs)

    return wrapper


def _setter(func):
    """Add a private attribute to the class (``self._is_set = True``)
    when a method with this decorator is used.

    .. note::
       The method must be used as a decorator.
    """

    def wrapper(self, *args, **kwargs):
        """Call the original function"""
        if hasattr(self, "_is_set"):
            self._is_set = True
        else:
            setattr(self, "_is_set", True)
        func(self, *args, **kwargs)

    return wrapper



"""
rom_data_provider
=================
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class rom_data_provider(Operator):
    """Set the required data for the invariant terms computation (reduced
    matrices, lumped mass matrix, modes ...)

    Parameters
    ----------
    rom_type : bool
        If this pin is set to true, customized rom
        data must be given
    reduced_stiff_matrix : FieldsContainer
        Fieldscontainers containing the reduced
        stiffness matrix
    reduced_damping_matrix : FieldsContainer
        Fieldscontainers containing the reduced mass
        matrix
    reduced_mass_matrix : FieldsContainer
        Fieldscontainers containing the reduced damp
        matrix
    data_sources : DataSources
    reduced_rhs_vector : FieldsContainer
        Fieldscontainers containing the reduced rhs
        vector
    lumped_mass_matrix : FieldsContainer
        Fieldscontainers containing the lumped mass
        matrix
    mode_shapes : FieldsContainer
        Fieldscontainers containing the customized
        mode shapes


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.result.rom_data_provider()

    >>> # Make input connections
    >>> my_rom_type = bool()
    >>> op.inputs.rom_type.connect(my_rom_type)
    >>> my_reduced_stiff_matrix = dpf.FieldsContainer()
    >>> op.inputs.reduced_stiff_matrix.connect(my_reduced_stiff_matrix)
    >>> my_reduced_damping_matrix = dpf.FieldsContainer()
    >>> op.inputs.reduced_damping_matrix.connect(my_reduced_damping_matrix)
    >>> my_reduced_mass_matrix = dpf.FieldsContainer()
    >>> op.inputs.reduced_mass_matrix.connect(my_reduced_mass_matrix)
    >>> my_data_sources = dpf.DataSources()
    >>> op.inputs.data_sources.connect(my_data_sources)
    >>> my_reduced_rhs_vector = dpf.FieldsContainer()
    >>> op.inputs.reduced_rhs_vector.connect(my_reduced_rhs_vector)
    >>> my_lumped_mass_matrix = dpf.FieldsContainer()
    >>> op.inputs.lumped_mass_matrix.connect(my_lumped_mass_matrix)
    >>> my_mode_shapes = dpf.FieldsContainer()
    >>> op.inputs.mode_shapes.connect(my_mode_shapes)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.result.rom_data_provider(
    ...     rom_type=my_rom_type,
    ...     reduced_stiff_matrix=my_reduced_stiff_matrix,
    ...     reduced_damping_matrix=my_reduced_damping_matrix,
    ...     reduced_mass_matrix=my_reduced_mass_matrix,
    ...     data_sources=my_data_sources,
    ...     reduced_rhs_vector=my_reduced_rhs_vector,
    ...     lumped_mass_matrix=my_lumped_mass_matrix,
    ...     mode_shapes=my_mode_shapes,
    ... )

    >>> # Get output data
    >>> result_rom_matrices = op.outputs.rom_matrices()
    >>> result_mode_shapes = op.outputs.mode_shapes()
    >>> result_lumped_mass = op.outputs.lumped_mass()
    >>> result_model_data = op.outputs.model_data()
    >>> result_center_of_mass = op.outputs.center_of_mass()
    >>> result_inertia_relief = op.outputs.inertia_relief()
    >>> result_model_size = op.outputs.model_size()
    >>> result_field_coordinates_and_euler_angles = op.outputs.field_coordinates_and_euler_angles()
    >>> result_nod = op.outputs.nod()
    """

    def __init__(
        self,
        rom_type=None,
        reduced_stiff_matrix=None,
        reduced_damping_matrix=None,
        reduced_mass_matrix=None,
        data_sources=None,
        reduced_rhs_vector=None,
        lumped_mass_matrix=None,
        mode_shapes=None,
        config=None,
        server=None,
    ):
        super().__init__(name="rom_data_provider", config=config, server=server)
        self._inputs = InputsRomDataProvider(self)
        self._outputs = OutputsRomDataProvider(self)
        if rom_type is not None:
            self.inputs.rom_type.connect(rom_type)
        if reduced_stiff_matrix is not None:
            self.inputs.reduced_stiff_matrix.connect(reduced_stiff_matrix)
        if reduced_damping_matrix is not None:
            self.inputs.reduced_damping_matrix.connect(reduced_damping_matrix)
        if reduced_mass_matrix is not None:
            self.inputs.reduced_mass_matrix.connect(reduced_mass_matrix)
        if data_sources is not None:
            self.inputs.data_sources.connect(data_sources)
        if reduced_rhs_vector is not None:
            self.inputs.reduced_rhs_vector.connect(reduced_rhs_vector)
        if lumped_mass_matrix is not None:
            self.inputs.lumped_mass_matrix.connect(lumped_mass_matrix)
        if mode_shapes is not None:
            self.inputs.mode_shapes.connect(mode_shapes)

    @staticmethod
    def _spec():
        description = """Set the required data for the invariant terms computation (reduced
            matrices, lumped mass matrix, modes ...)"""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="rom_type",
                    type_names=["bool"],
                    optional=False,
                    document="""If this pin is set to true, customized rom
        data must be given""",
                ),
                1: PinSpecification(
                    name="reduced_stiff_matrix",
                    type_names=["fields_container"],
                    optional=False,
                    document="""Fieldscontainers containing the reduced
        stiffness matrix""",
                ),
                2: PinSpecification(
                    name="reduced_damping_matrix",
                    type_names=["fields_container"],
                    optional=False,
                    document="""Fieldscontainers containing the reduced mass
        matrix""",
                ),
                3: PinSpecification(
                    name="reduced_mass_matrix",
                    type_names=["fields_container"],
                    optional=False,
                    document="""Fieldscontainers containing the reduced damp
        matrix""",
                ),
                4: PinSpecification(
                    name="data_sources",
                    type_names=["data_sources"],
                    optional=False,
                    document="""""",
                ),
                5: PinSpecification(
                    name="reduced_rhs_vector",
                    type_names=["fields_container"],
                    optional=False,
                    document="""Fieldscontainers containing the reduced rhs
        vector""",
                ),
                6: PinSpecification(
                    name="lumped_mass_matrix",
                    type_names=["fields_container"],
                    optional=False,
                    document="""Fieldscontainers containing the lumped mass
        matrix""",
                ),
                7: PinSpecification(
                    name="mode_shapes",
                    type_names=["fields_container"],
                    optional=False,
                    document="""Fieldscontainers containing the customized
        mode shapes""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="rom_matrices",
                    type_names=["fields_container"],
                    optional=False,
                    document="""Fieldscontainers containing the reduced
        matrices""",
                ),
                1: PinSpecification(
                    name="mode_shapes",
                    type_names=["fields_container"],
                    optional=False,
                    document="""Fieldscontainers containing the mode shapes,
        which are cst and nor for the cms
        method""",
                ),
                2: PinSpecification(
                    name="lumped_mass",
                    type_names=["fields_container"],
                    optional=False,
                    document="""Fieldscontainers containing the lumped mass""",
                ),
                3: PinSpecification(
                    name="model_data",
                    type_names=["property_field"],
                    optional=False,
                    document="""Data describing the finite element model""",
                ),
                4: PinSpecification(
                    name="center_of_mass",
                    type_names=["property_field"],
                    optional=False,
                    document="""""",
                ),
                5: PinSpecification(
                    name="inertia_relief",
                    type_names=["field"],
                    optional=False,
                    document="""Inertia matrix""",
                ),
                6: PinSpecification(
                    name="model_size",
                    type_names=["double"],
                    optional=False,
                    document="""Size of the model""",
                ),
                7: PinSpecification(
                    name="field_coordinates_and_euler_angles",
                    type_names=["double"],
                    optional=False,
                    document="""Coordinates and euler angles of all nodes""",
                ),
                8: PinSpecification(
                    name="nod",
                    type_names=["vector<int32>"],
                    optional=False,
                    document="""Ids of master nodes""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="rom_data_provider", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsRomDataProvider
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsRomDataProvider
        """
        return super().outputs


class InputsRomDataProvider(_Inputs):
    """Intermediate class used to connect user inputs to
    rom_data_provider operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.result.rom_data_provider()
    >>> my_rom_type = bool()
    >>> op.inputs.rom_type.connect(my_rom_type)
    >>> my_reduced_stiff_matrix = dpf.FieldsContainer()
    >>> op.inputs.reduced_stiff_matrix.connect(my_reduced_stiff_matrix)
    >>> my_reduced_damping_matrix = dpf.FieldsContainer()
    >>> op.inputs.reduced_damping_matrix.connect(my_reduced_damping_matrix)
    >>> my_reduced_mass_matrix = dpf.FieldsContainer()
    >>> op.inputs.reduced_mass_matrix.connect(my_reduced_mass_matrix)
    >>> my_data_sources = dpf.DataSources()
    >>> op.inputs.data_sources.connect(my_data_sources)
    >>> my_reduced_rhs_vector = dpf.FieldsContainer()
    >>> op.inputs.reduced_rhs_vector.connect(my_reduced_rhs_vector)
    >>> my_lumped_mass_matrix = dpf.FieldsContainer()
    >>> op.inputs.lumped_mass_matrix.connect(my_lumped_mass_matrix)
    >>> my_mode_shapes = dpf.FieldsContainer()
    >>> op.inputs.mode_shapes.connect(my_mode_shapes)
    """

    def __init__(self, op: Operator):
        super().__init__(rom_data_provider._spec().inputs, op)
        self._rom_type = Input(rom_data_provider._spec().input_pin(0), 0, op, -1)
        self._inputs.append(self._rom_type)
        self._reduced_stiff_matrix = Input(
            rom_data_provider._spec().input_pin(1), 1, op, -1
        )
        self._inputs.append(self._reduced_stiff_matrix)
        self._reduced_damping_matrix = Input(
            rom_data_provider._spec().input_pin(2), 2, op, -1
        )
        self._inputs.append(self._reduced_damping_matrix)
        self._reduced_mass_matrix = Input(
            rom_data_provider._spec().input_pin(3), 3, op, -1
        )
        self._inputs.append(self._reduced_mass_matrix)
        self._data_sources = Input(rom_data_provider._spec().input_pin(4), 4, op, -1)
        self._inputs.append(self._data_sources)
        self._reduced_rhs_vector = Input(
            rom_data_provider._spec().input_pin(5), 5, op, -1
        )
        self._inputs.append(self._reduced_rhs_vector)
        self._lumped_mass_matrix = Input(
            rom_data_provider._spec().input_pin(6), 6, op, -1
        )
        self._inputs.append(self._lumped_mass_matrix)
        self._mode_shapes = Input(rom_data_provider._spec().input_pin(7), 7, op, -1)
        self._inputs.append(self._mode_shapes)

    @property
    def rom_type(self):
        """Allows to connect rom_type input to the operator.

        If this pin is set to true, customized rom
        data must be given

        Parameters
        ----------
        my_rom_type : bool

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.rom_data_provider()
        >>> op.inputs.rom_type.connect(my_rom_type)
        >>> # or
        >>> op.inputs.rom_type(my_rom_type)
        """
        return self._rom_type

    @property
    def reduced_stiff_matrix(self):
        """Allows to connect reduced_stiff_matrix input to the operator.

        Fieldscontainers containing the reduced
        stiffness matrix

        Parameters
        ----------
        my_reduced_stiff_matrix : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.rom_data_provider()
        >>> op.inputs.reduced_stiff_matrix.connect(my_reduced_stiff_matrix)
        >>> # or
        >>> op.inputs.reduced_stiff_matrix(my_reduced_stiff_matrix)
        """
        return self._reduced_stiff_matrix

    @property
    def reduced_damping_matrix(self):
        """Allows to connect reduced_damping_matrix input to the operator.

        Fieldscontainers containing the reduced mass
        matrix

        Parameters
        ----------
        my_reduced_damping_matrix : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.rom_data_provider()
        >>> op.inputs.reduced_damping_matrix.connect(my_reduced_damping_matrix)
        >>> # or
        >>> op.inputs.reduced_damping_matrix(my_reduced_damping_matrix)
        """
        return self._reduced_damping_matrix

    @property
    def reduced_mass_matrix(self):
        """Allows to connect reduced_mass_matrix input to the operator.

        Fieldscontainers containing the reduced damp
        matrix

        Parameters
        ----------
        my_reduced_mass_matrix : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.rom_data_provider()
        >>> op.inputs.reduced_mass_matrix.connect(my_reduced_mass_matrix)
        >>> # or
        >>> op.inputs.reduced_mass_matrix(my_reduced_mass_matrix)
        """
        return self._reduced_mass_matrix

    @property
    def data_sources(self):
        """Allows to connect data_sources input to the operator.

        Parameters
        ----------
        my_data_sources : DataSources

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.rom_data_provider()
        >>> op.inputs.data_sources.connect(my_data_sources)
        >>> # or
        >>> op.inputs.data_sources(my_data_sources)
        """
        return self._data_sources

    @property
    def reduced_rhs_vector(self):
        """Allows to connect reduced_rhs_vector input to the operator.

        Fieldscontainers containing the reduced rhs
        vector

        Parameters
        ----------
        my_reduced_rhs_vector : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.rom_data_provider()
        >>> op.inputs.reduced_rhs_vector.connect(my_reduced_rhs_vector)
        >>> # or
        >>> op.inputs.reduced_rhs_vector(my_reduced_rhs_vector)
        """
        return self._reduced_rhs_vector

    @property
    def lumped_mass_matrix(self):
        """Allows to connect lumped_mass_matrix input to the operator.

        Fieldscontainers containing the lumped mass
        matrix

        Parameters
        ----------
        my_lumped_mass_matrix : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.rom_data_provider()
        >>> op.inputs.lumped_mass_matrix.connect(my_lumped_mass_matrix)
        >>> # or
        >>> op.inputs.lumped_mass_matrix(my_lumped_mass_matrix)
        """
        return self._lumped_mass_matrix

    @property
    def mode_shapes(self):
        """Allows to connect mode_shapes input to the operator.

        Fieldscontainers containing the customized
        mode shapes

        Parameters
        ----------
        my_mode_shapes : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.rom_data_provider()
        >>> op.inputs.mode_shapes.connect(my_mode_shapes)
        >>> # or
        >>> op.inputs.mode_shapes(my_mode_shapes)
        """
        return self._mode_shapes


class OutputsRomDataProvider(_Outputs):
    """Intermediate class used to get outputs from
    rom_data_provider operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.result.rom_data_provider()
    >>> # Connect inputs : op.inputs. ...
    >>> result_rom_matrices = op.outputs.rom_matrices()
    >>> result_mode_shapes = op.outputs.mode_shapes()
    >>> result_lumped_mass = op.outputs.lumped_mass()
    >>> result_model_data = op.outputs.model_data()
    >>> result_center_of_mass = op.outputs.center_of_mass()
    >>> result_inertia_relief = op.outputs.inertia_relief()
    >>> result_model_size = op.outputs.model_size()
    >>> result_field_coordinates_and_euler_angles = op.outputs.field_coordinates_and_euler_angles()
    >>> result_nod = op.outputs.nod()
    """

    def __init__(self, op: Operator):
        super().__init__(rom_data_provider._spec().outputs, op)
        self._rom_matrices = Output(rom_data_provider._spec().output_pin(0), 0, op)
        self._outputs.append(self._rom_matrices)
        self._mode_shapes = Output(rom_data_provider._spec().output_pin(1), 1, op)
        self._outputs.append(self._mode_shapes)
        self._lumped_mass = Output(rom_data_provider._spec().output_pin(2), 2, op)
        self._outputs.append(self._lumped_mass)
        self._model_data = Output(rom_data_provider._spec().output_pin(3), 3, op)
        self._outputs.append(self._model_data)
        self._center_of_mass = Output(rom_data_provider._spec().output_pin(4), 4, op)
        self._outputs.append(self._center_of_mass)
        self._inertia_relief = Output(rom_data_provider._spec().output_pin(5), 5, op)
        self._outputs.append(self._inertia_relief)
        self._model_size = Output(rom_data_provider._spec().output_pin(6), 6, op)
        self._outputs.append(self._model_size)
        self._field_coordinates_and_euler_angles = Output(
            rom_data_provider._spec().output_pin(7), 7, op
        )
        self._outputs.append(self._field_coordinates_and_euler_angles)
        self._nod = Output(rom_data_provider._spec().output_pin(8), 8, op)
        self._outputs.append(self._nod)

    @property
    def rom_matrices(self):
        """Allows to get rom_matrices output of the operator

        Returns
        ----------
        my_rom_matrices : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.rom_data_provider()
        >>> # Connect inputs : op.inputs. ...
        >>> result_rom_matrices = op.outputs.rom_matrices()
        """  # noqa: E501
        return self._rom_matrices

    @property
    def mode_shapes(self):
        """Allows to get mode_shapes output of the operator

        Returns
        ----------
        my_mode_shapes : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.rom_data_provider()
        >>> # Connect inputs : op.inputs. ...
        >>> result_mode_shapes = op.outputs.mode_shapes()
        """  # noqa: E501
        return self._mode_shapes

    @property
    def lumped_mass(self):
        """Allows to get lumped_mass output of the operator

        Returns
        ----------
        my_lumped_mass : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.rom_data_provider()
        >>> # Connect inputs : op.inputs. ...
        >>> result_lumped_mass = op.outputs.lumped_mass()
        """  # noqa: E501
        return self._lumped_mass

    @property
    def model_data(self):
        """Allows to get model_data output of the operator

        Returns
        ----------
        my_model_data : PropertyField

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.rom_data_provider()
        >>> # Connect inputs : op.inputs. ...
        >>> result_model_data = op.outputs.model_data()
        """  # noqa: E501
        return self._model_data

    @property
    def center_of_mass(self):
        """Allows to get center_of_mass output of the operator

        Returns
        ----------
        my_center_of_mass : PropertyField

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.rom_data_provider()
        >>> # Connect inputs : op.inputs. ...
        >>> result_center_of_mass = op.outputs.center_of_mass()
        """  # noqa: E501
        return self._center_of_mass

    @property
    def inertia_relief(self):
        """Allows to get inertia_relief output of the operator

        Returns
        ----------
        my_inertia_relief : Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.rom_data_provider()
        >>> # Connect inputs : op.inputs. ...
        >>> result_inertia_relief = op.outputs.inertia_relief()
        """  # noqa: E501
        return self._inertia_relief

    @property
    def model_size(self):
        """Allows to get model_size output of the operator

        Returns
        ----------
        my_model_size : float

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.rom_data_provider()
        >>> # Connect inputs : op.inputs. ...
        >>> result_model_size = op.outputs.model_size()
        """  # noqa: E501
        return self._model_size

    @property
    def field_coordinates_and_euler_angles(self):
        """Allows to get field_coordinates_and_euler_angles output of the operator

        Returns
        ----------
        my_field_coordinates_and_euler_angles : float

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.rom_data_provider()
        >>> # Connect inputs : op.inputs. ...
        >>> result_field_coordinates_and_euler_angles = op.outputs.field_coordinates_and_euler_angles()
        """  # noqa: E501
        return self._field_coordinates_and_euler_angles

    @property
    def nod(self):
        """Allows to get nod output of the operator

        Returns
        ----------
        my_nod :

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.rom_data_provider()
        >>> # Connect inputs : op.inputs. ...
        >>> result_nod = op.outputs.nod()
        """  # noqa: E501
        return self._nod



"""
MeshedRegion
============
"""
import traceback
import warnings

import ansys.dpf.core.errors

from ansys.dpf.core import scoping, field, property_field
from ansys.dpf.core.check_version import server_meet_version, version_requires
from ansys.dpf.core.common import (
    locations,
    types,
    nodal_properties,
)
from ansys.dpf.core.elements import Elements, element_types
from ansys.dpf.core.nodes import Nodes
from ansys.dpf.core.faces import Faces
from ansys.dpf.core.plotter import DpfPlotter, Plotter
from ansys.dpf.core.cache import class_handling_cache
from ansys.dpf.core import server as server_module
from ansys.dpf.gate import meshed_region_capi, meshed_region_grpcapi


def update_grid(func):
    # Decorate mesh setters to centralize the update logic of pyvista objects.
    def wrapper(*args, **kwargs):
        mesh = args[0]
        if mesh._full_grid is not None:
            # Treat each setter separately to improve performance by updating the minimum required.
            if func.__name__ == "set_coordinates_field":
                # When setting node coordinates
                from ansys.dpf.core.vtk_helper import vtk_update_coordinates

                vtk_update_coordinates(vtk_grid=mesh._full_grid, coordinates_array=args[1].data)

        return func(*args, **kwargs)

    return wrapper


@class_handling_cache
class MeshedRegion:
    """
    Represents a mesh from DPF.

    Parameters
    ----------
    num_nodes : int, optional
        Number of nodes to reserve for mesh creation. The default is ``None``.
    num_elements : int, optional
        Number of elements to reserve for mesh creation. The default is ``None``.
    mesh : ansys.grpc.dpf.meshed_region_pb2.MeshedRegion
        The default is ``None``.
    server : ansys.dpf.core.server, optional
        Server with the channel connected to the remote or local instance.
        The default is ``None``, in which case an attempt is made to use the
        global server.

    Examples
    --------
    Extract a meshed region from a model.

    >>> import ansys.dpf.core as dpf
    >>> from ansys.dpf.core import examples
    >>> model = dpf.Model(examples.find_static_rst())
    >>> meshed_region = model.metadata.meshed_region

    Create a meshed region from scratch (line with 3 beam elements).

    >>> import ansys.dpf.core as dpf
    >>> meshed_region = dpf.MeshedRegion(num_nodes=4,num_elements=3)
    >>> i=0
    >>> for node in meshed_region.nodes.add_nodes(4):
    ...     node.id = i+1
    ...     node.coordinates = [float(i), float(i), 0.0]
    ...     i=i+1
    >>> i=0
    >>> for element in meshed_region.elements.add_elements(3):
    ...     element.id=i+1
    ...     element.connectivity = [i, i+1]
    ...     element.is_beam=True #or is_solid, is_beam, is_point
    ...     i=i+1
    >>> meshed_region.elements.add_beam_element(id=4,connectivity=[3,0])

    """

    def __init__(self, num_nodes=None, num_elements=None, mesh=None, server=None):
        # step 1: get server
        self._server = server_module.get_or_create_server(server)

        # step 2: get api
        self._api = self._server.get_api_for_type(
            capi=meshed_region_capi.MeshedRegionCAPI,
            grpcapi=meshed_region_grpcapi.MeshedRegionGRPCAPI,
        )

        # step3: init environment
        self._api.init_meshed_region_environment(self)  # creates stub when gRPC

        # step4: if object exists: take instance, else create it:
        # object_name -> protobuf.message, DPFObject*
        if mesh is not None:
            self._internal_obj = mesh
        else:
            # if no mesh object, create one
            if self._server.has_client():
                self._internal_obj = self._api.meshed_region_new_on_client(self._server.client)
            else:
                self._internal_obj = self._api.meshed_region_new()

        self._full_grid = None
        self._elements = None
        self._nodes = None
        self.as_linear = None

    def _get_scoping(self, loc=locations.nodal):
        """
        Parameters
        ----------
        loc : str or ansys.dpf.core.common.locations, optional
            location of the requested scoping ("Nodal", "Elemental"...)

        Returns
        -------
        scoping : Scoping
            ids of the elements or nodes of the mesh
        """
        if loc is locations.nodal:
            out = self._api.meshed_region_get_shared_nodes_scoping(self)
        elif loc is locations.elemental:
            out = self._api.meshed_region_get_shared_elements_scoping(self)
        else:
            raise TypeError(f"Location {loc} is not recognized.")
        if out is None:
            return
        scop_to_return = scoping.Scoping(scoping=out, server=self._server)
        try:
            check = scop_to_return._api.scoping_fast_access_ptr(scop_to_return)
            if check is None:
                return None
        except NotImplementedError:
            # will throw NotImplementedError for ansys-grpc-dpf
            pass
        except ansys.dpf.core.errors.DPFServerException:
            # and DPFServerException for gRPC CLayer
            pass
        return scop_to_return

    @property
    def elements(self):
        """
        All elemental properties of the mesh, such as connectivity and element types.

        Returns
        -------
        elements : Elements
            Elements belonging to the meshed region.

        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.find_static_rst())
        >>> meshed_region = model.metadata.meshed_region
        >>> elements = meshed_region.elements
        >>> print(elements)
        DPF Elements object with 8 elements

        """
        return Elements(self)

    @property
    def faces(self):
        """
        All face properties of the mesh, such as faces_nodes_connectivity and face types.

        Returns
        -------
        faces : Faces
            Faces belonging to the meshed region.

        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.find_static_rst())
        >>> meshed_region = model.metadata.meshed_region
        >>> faces = meshed_region.faces
        >>> print(faces)
        DPF Faces object with 0 faces

        """
        return Faces(self)

    @property
    def nodes(self):
        """
        All nodal properties of the mesh, such as node coordinates and nodal connectivity.

        Returns
        -------
        nodes : Nodes
            Nodes belonging to the meshed region

        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.find_static_rst())
        >>> meshed_region = model.metadata.meshed_region
        >>> nodes = meshed_region.nodes
        >>> nodes.n_nodes
        81

        """
        return Nodes(self)

    @property
    def unit(self):
        """
        Unit of the meshed region.

        This unit is the same as the unit of the coordinates of the meshed region.

        Returns
        -------
        unit : str
        """
        return self._get_unit()

    @unit.setter
    def unit(self, value):
        """
        Unit type.

        Parameters
        ----------
        unit : str
        """
        return self._set_unit(value)

    def _get_unit(self):
        """
        Retrieve the unit type.

        Returns
        -------
        unit : str
        """
        return self._api.meshed_region_get_unit(self)

    def _set_unit(self, unit):
        """
        Set the unit of the meshed region.

        Parameters
        ----------
        unit: str
        """
        return self._api.meshed_region_set_unit(self, unit)

    def __del__(self):
        try:
            self._deleter_func[0](self._deleter_func[1](self))
        except:
            warnings.warn(traceback.format_exc())

    def __str__(self):
        from ansys.dpf.core.core import _description

        return _description(self._internal_obj, self._server)

    @property
    def available_property_fields(self):
        """
        Returns a list of available property fields

        Returns
        -------
        available_property_fields : list str
        """
        available_property_fields = []
        n_property_field = self._api.meshed_region_get_num_available_property_field(self)
        for index in range(n_property_field):
            available_property_fields.append(
                self._api.meshed_region_get_property_field_name(self, index)
            )
        return available_property_fields

    def property_field(self, property_name):
        """
        Property field getter. It can be coordinates (field),
        element types (property field)...

        Returns
        -------
        field_or_property_field : core.Field or core.PropertyField
        """
        return self.field_of_properties(property_name)

    @version_requires("3.0")
    def set_property_field(self, property_name, value):
        """
        Property field setter. It can be coordinates (field),
        element types (property field)...

        Parameters
        ----------
        property_name : str
            property name of the field to set
        value : PropertyField or Field
        """
        if property_name is nodal_properties.coordinates:
            self.set_coordinates_field(value)
        else:
            self._api.meshed_region_set_property_field(self, property_name, value)

    @update_grid
    @version_requires("3.0")
    def set_coordinates_field(self, coordinates_field):
        """
        Coordinates field setter.

        Parameters
        ----------
        coordinates_field : PropertyField or Field
        """
        self._api.meshed_region_set_coordinates_field(self, coordinates_field)

    @property
    def available_named_selections(self):
        """
        List of available named selections.

        Returns
        -------
        named_selections : list str
        """
        return self._get_available_named_selections()

    def _get_available_named_selections(self):
        """
        List of available named selections.

        Returns
        -------
        named_selections : list str
        """
        named_selections = []
        n_selections = self._api.meshed_region_get_num_available_named_selection(self)
        for index in range(n_selections):
            named_selections.append(self._api.meshed_region_get_named_selection_name(self, index))
        return named_selections

    def named_selection(self, named_selection):
        """
        Scoping containing the list of nodes or elements in the named selection.

        Parameters
        ----------
        named_selection : str
            Name of the named selection.

        Returns
        -------
        named_selection : Scoping
        """
        if server_meet_version("2.1", self._server):
            out = self._api.meshed_region_get_named_selection_scoping(self, named_selection)
            return scoping.Scoping(scoping=out, server=self._server)
        else:
            if hasattr(self, "_stream_provider"):
                from ansys.dpf.core.dpf_operator import Operator

                op = Operator("scoping_provider_by_ns", server=self._server)
                op.connect(1, named_selection)
                op.connect(3, self._stream_provider, 0)
                return op.get_output(0, types.scoping)
            else:
                raise Exception(
                    "Getting a named selection from a meshed region is "
                    "only implemented for meshed region created from a "
                    "model for server version 2.0. Please update your server."
                )

    @version_requires("3.0")
    def set_named_selection_scoping(self, named_selection_name, scoping):
        """
        Named selection scoping setter.

        Parameters
        ----------
        named_selection_name : str
            named selection name
        scoping : Scoping
        """
        return self._api.meshed_region_set_named_selection_scoping(
            self, named_selection_name, scoping
        )

    def _set_stream_provider(self, stream_provider):
        self._stream_provider = stream_provider

    # NOTE: kept only for reference as the mesh operator is being moved out of dpf
    # def write_vtk(self, filename, skin_only=True):
    #     """Return a vtk mesh"""
    #     # filename = os.path.join(tempfile.gettempdir(),
    #                             # '%s.vtk' % next(tempfile._get_candidate_names()))

    #     vtk_exp = self._model.operator("vtk_export")
    #     vtk_exp.connect(0, filename)

    #     mesh = self._model.operator("mapdl::rst::MeshProvider")
    #     mesh.connect(4, self._model.data_sources)

    #     if skin_only:
    #         skin = self._model.operator("meshed_skin_sector")
    #         skin.connect(0, mesh, 0)
    #         vtk_exp.connect(1, skin, 0)
    #     else:
    #         vtk_exp.connect(1, mesh, 0)

    #     vtk_exp.run()
    #     if not os.path.isfile(filename):
    #         raise FileNotFoundError('VTK mesh not written to disk')

    # @property
    # def skin(self):
    #     """Surface of the meshed region."""
    #     mesh = self._model.operator("mapdl::rst::MeshProvider")
    #     mesh.connect(4, self._model.data_sources)

    #     skin = self._model.operator("meshed_skin_sector")
    #     skin.connect(0, mesh, 0)
    #     # skin.connect(4, self)

    #     skin.get_output(0, types.meshed_region)

    #     name = None
    #     if self._name:
    #         name = 'Skin of %s' % self._name
    #     self._internal_obj = skin.get_output(0, types.meshed_region)
    #     return MeshedRegion(self._server.channel, skin, self._model, name)

    def deform_by(self, deform_by, scale_factor=1.0):
        """
        Deforms the mesh according to a 3D vector field and an additional scale factor.

        Parameters
        ----------
        deform_by : Field, FieldsContainer, Result, Operator
            Used to deform the plotted mesh. Must output a unique 3D vector field.
            Defaults to None.
        scale_factor : float, Field, FieldsContainer, optional
            Used to scale the mesh deformation. Defaults to 1.0. Can be a scalar Field
            (or a FieldsContainer with only one Field) to get a spatially non-homogeneous scaling.
        Returns
        -------

        """
        from ansys.dpf.core.operators.math import add, scale, unit_convert

        if hasattr(deform_by, "eval"):
            # If a Result or an Operator, eval and get the field.
            deform_by = deform_by.eval()[0]
        if isinstance(deform_by, ansys.dpf.core.fields_container.FieldsContainer):
            # If a FieldsContainer, take the field.
            deform_by = deform_by[0]
        if deform_by.unit != self.unit:
            unit_convert(deform_by, self.unit)
        scale_op = scale(field=deform_by, ponderation=scale_factor)
        return add(fieldA=self.nodes.coordinates_field, fieldB=scale_op.outputs.field).eval()

    def _as_vtk(self, coordinates=None, as_linear=True, include_ids=False):
        """Convert DPF mesh to a PyVista unstructured grid."""
        try:
            from ansys.dpf.core.vtk_helper import dpf_mesh_to_vtk
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "To use plotting capabilities, please install pyvista "
                "with :\n pip install pyvista>=0.24.0"
            )
        grid = dpf_mesh_to_vtk(self, coordinates, as_linear)

        # consider adding this when scoping request is faster
        if include_ids:
            self._nodeids = self.nodes.scoping.ids
            self._elementids = self.elements.scoping.ids
            grid["node_ids"] = self._nodeids
            grid["element_ids"] = self._elementids

        self.as_linear = as_linear  # store as_linear to avoid passing through here again
        return grid

    @property
    def grid(self):
        """
        Unstructured grid in VTK format from PyVista.

        Returns
        -------
        pyvista.UnstructuredGrid
            UnstructuredGrid of the mesh.

        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.find_static_rst())
        >>> meshed_region = model.metadata.meshed_region
        >>> grid = meshed_region.grid

        Plot this grid directly.

        >>> grid.plot()

        Extract the surface mesh of this grid

        >>> mesh = grid.extract_surface()

        """
        if self._full_grid is None:
            self._full_grid = self._as_vtk(self.nodes.coordinates_field)
        return self._full_grid

    def plot(
        self,
        field_or_fields_container=None,
        shell_layers=None,
        deform_by=None,
        scale_factor=1.0,
        **kwargs,
    ):
        """
        Plot the field or fields container on the mesh.

        Parameters
        ----------
        field_or_fields_container : dpf.core.Field or dpf.core.FieldsContainer
            Field or fields container to plot. The default is ``None``.
        shell_layers : core.shell_layers, optional
            Enum used to set the shell layers if the model to plot contains shell elements.
        deform_by : Field, Result, Operator, optional
            Used to deform the plotted mesh. Must output a 3D vector field.
            Defaults to None.
        scale_factor : float, optional
            Scaling factor to apply when warping the mesh. Defaults to 1.0.
        **kwargs : optional
            Additional keyword arguments for the plotter. For additional keyword
            arguments, see ``help(pyvista.plot)``.

        Examples
        --------
        Plot the displacement field from an example file.

        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.find_static_rst())
        >>> disp = model.results.displacement()
        >>> field = disp.outputs.fields_container()[0]
        >>> model.metadata.meshed_region.plot(field)

        """
        if field_or_fields_container is not None:
            pl = Plotter(self, **kwargs)
            return pl.plot_contour(
                field_or_fields_container,
                shell_layers,
                show_axes=kwargs.pop("show_axes", True),
                deform_by=deform_by,
                scale_factor=scale_factor,
                **kwargs,
            )

        # otherwise, simply plot the mesh
        pl = DpfPlotter(**kwargs)
        pl.add_mesh(
            self,
            deform_by=deform_by,
            scale_factor=scale_factor,
            show_axes=kwargs.pop("show_axes", True),
            **kwargs,
        )
        return pl.show_figure(**kwargs)

    def deep_copy(self, server=None):
        """
        Create a deep copy of the meshed region's data on a given server.

        This method is useful for passing data from one server instance to another.

        Parameters
        ----------
        server : ansys.dpf.core.server, optional
            Server with the channel connected to the remote or local instance.
            The default is ``None``, in which case an attempt is made to use the
            global server.

        Returns
        -------
        mesh_copy : MeshedRegion

        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.find_static_rst())
        >>> meshed_region = model.metadata.meshed_region
        >>> other_server = dpf.start_local_server(as_global=False)
        >>> deep_copy = meshed_region.deep_copy(server=other_server)

        """
        if self._server.config.legacy:
            if self.nodes.scoping is None:  # empty Mesh
                return MeshedRegion()
            node_ids = self.nodes.scoping.ids
            element_ids = self.elements.scoping.ids
            mesh = MeshedRegion(
                num_nodes=len(node_ids), num_elements=len(element_ids), server=server
            )
            with self.nodes.coordinates_field.as_local_field() as coord:
                for i, node in enumerate(mesh.nodes.add_nodes(len(node_ids))):
                    node.id = node_ids[i]
                    node.coordinates = coord.get_entity_data(i)
            with self.elements.connectivities_field.as_local_field() as connect:
                with self.elements.element_types_field.as_local_field() as types:
                    for i, elem in enumerate(mesh.elements.add_elements(len(element_ids))):
                        elem.id = element_ids[i]
                        elem.connectivity = connect.get_entity_data(i)
                        elem.shape = element_types.shape(types.get_entity_data(i)[0])
            mesh.unit = self.unit
            return mesh
        else:
            from ansys.dpf.core.core import _deep_copy

            return _deep_copy(self, server=server)

    def field_of_properties(self, property_name):
        """
        Returns the ``Field`` or ``PropertyField`` associated
        to a given property of the mesh

        Parameters
        ----------
        property_name : str, common.elemental_properties, common.nodal_properties
            Name of the property.

        Returns
        -------
        properties : Field, PropertyField

        Examples
        --------
        >>> import ansys.dpf.core as dpf
        >>> from ansys.dpf.core import examples
        >>> model = dpf.Model(examples.find_static_rst())
        >>> meshed_region = model.metadata.meshed_region
        >>> connectivity = meshed_region.field_of_properties(
        ...     dpf.common.elemental_properties.connectivity)
        >>> coordinates = meshed_region.field_of_properties(dpf.common.nodal_properties.coordinates)
        """
        if property_name is nodal_properties.coordinates:
            field_out = self._api.meshed_region_get_coordinates_field(self)
            return field.Field(server=self._server, field=field_out)
        else:
            field_out = self._api.meshed_region_get_property_field(self, property_name)
            if isinstance(field_out, int):
                res = property_field.PropertyField(server=self._server, property_field=field_out)
                return res
            else:
                if field_out.datatype == "int":
                    return property_field.PropertyField(
                        server=self._server, property_field=field_out
                    )
                else:
                    # Not sure we go through here since the only datatype not int is coordinates,
                    # which is already dealt with previously.
                    return field.Field(server=self._server, field=field_out)



"""
compute_stress_Z
================
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class compute_stress_Z(Operator):
    """Computes the stress from an elastic strain field.compute_total_strain
    limitations are applicable for stress computationGet the ZZ normal
    component (22 component).

    Parameters
    ----------
    scoping : Scoping, optional
        The element scoping on which the result is
        computed.
    streams_container : StreamsContainer, optional
        Needed to get mesh and material ids. optional
        if a data_sources have been
        connected.
    data_sources : DataSources, optional
        Needed to get mesh and material ids. optional
        if a streams_container have been
        connected.
    requested_location : str, optional
        Average the elemental nodal result to the
        requested location.
    strain : FieldsContainer or Field
        Field/or fields container containing only the
        elastic strain field (element nodal).


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.result.compute_stress_Z()

    >>> # Make input connections
    >>> my_scoping = dpf.Scoping()
    >>> op.inputs.scoping.connect(my_scoping)
    >>> my_streams_container = dpf.StreamsContainer()
    >>> op.inputs.streams_container.connect(my_streams_container)
    >>> my_data_sources = dpf.DataSources()
    >>> op.inputs.data_sources.connect(my_data_sources)
    >>> my_requested_location = str()
    >>> op.inputs.requested_location.connect(my_requested_location)
    >>> my_strain = dpf.FieldsContainer()
    >>> op.inputs.strain.connect(my_strain)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.result.compute_stress_Z(
    ...     scoping=my_scoping,
    ...     streams_container=my_streams_container,
    ...     data_sources=my_data_sources,
    ...     requested_location=my_requested_location,
    ...     strain=my_strain,
    ... )

    >>> # Get output data
    >>> result_fields_container = op.outputs.fields_container()
    """

    def __init__(
        self,
        scoping=None,
        streams_container=None,
        data_sources=None,
        requested_location=None,
        strain=None,
        config=None,
        server=None,
    ):
        super().__init__(name="compute_stress_Z", config=config, server=server)
        self._inputs = InputsComputeStressZ(self)
        self._outputs = OutputsComputeStressZ(self)
        if scoping is not None:
            self.inputs.scoping.connect(scoping)
        if streams_container is not None:
            self.inputs.streams_container.connect(streams_container)
        if data_sources is not None:
            self.inputs.data_sources.connect(data_sources)
        if requested_location is not None:
            self.inputs.requested_location.connect(requested_location)
        if strain is not None:
            self.inputs.strain.connect(strain)

    @staticmethod
    def _spec():
        description = """Computes the stress from an elastic strain field.compute_total_strain
            limitations are applicable for stress computationGet the
            ZZ normal component (22 component)."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                1: PinSpecification(
                    name="scoping",
                    type_names=["scoping"],
                    optional=True,
                    document="""The element scoping on which the result is
        computed.""",
                ),
                3: PinSpecification(
                    name="streams_container",
                    type_names=["streams_container"],
                    optional=True,
                    document="""Needed to get mesh and material ids. optional
        if a data_sources have been
        connected.""",
                ),
                4: PinSpecification(
                    name="data_sources",
                    type_names=["data_sources"],
                    optional=True,
                    document="""Needed to get mesh and material ids. optional
        if a streams_container have been
        connected.""",
                ),
                9: PinSpecification(
                    name="requested_location",
                    type_names=["string"],
                    optional=True,
                    document="""Average the elemental nodal result to the
        requested location.""",
                ),
                10: PinSpecification(
                    name="strain",
                    type_names=["fields_container", "field"],
                    optional=False,
                    document="""Field/or fields container containing only the
        elastic strain field (element nodal).""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="fields_container",
                    type_names=["fields_container"],
                    optional=False,
                    document="""The computed result fields container
        (elemental nodal).""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="compute_stress_Z", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsComputeStressZ
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsComputeStressZ
        """
        return super().outputs


class InputsComputeStressZ(_Inputs):
    """Intermediate class used to connect user inputs to
    compute_stress_Z operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.result.compute_stress_Z()
    >>> my_scoping = dpf.Scoping()
    >>> op.inputs.scoping.connect(my_scoping)
    >>> my_streams_container = dpf.StreamsContainer()
    >>> op.inputs.streams_container.connect(my_streams_container)
    >>> my_data_sources = dpf.DataSources()
    >>> op.inputs.data_sources.connect(my_data_sources)
    >>> my_requested_location = str()
    >>> op.inputs.requested_location.connect(my_requested_location)
    >>> my_strain = dpf.FieldsContainer()
    >>> op.inputs.strain.connect(my_strain)
    """

    def __init__(self, op: Operator):
        super().__init__(compute_stress_Z._spec().inputs, op)
        self._scoping = Input(compute_stress_Z._spec().input_pin(1), 1, op, -1)
        self._inputs.append(self._scoping)
        self._streams_container = Input(
            compute_stress_Z._spec().input_pin(3), 3, op, -1
        )
        self._inputs.append(self._streams_container)
        self._data_sources = Input(compute_stress_Z._spec().input_pin(4), 4, op, -1)
        self._inputs.append(self._data_sources)
        self._requested_location = Input(
            compute_stress_Z._spec().input_pin(9), 9, op, -1
        )
        self._inputs.append(self._requested_location)
        self._strain = Input(compute_stress_Z._spec().input_pin(10), 10, op, -1)
        self._inputs.append(self._strain)

    @property
    def scoping(self):
        """Allows to connect scoping input to the operator.

        The element scoping on which the result is
        computed.

        Parameters
        ----------
        my_scoping : Scoping

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.compute_stress_Z()
        >>> op.inputs.scoping.connect(my_scoping)
        >>> # or
        >>> op.inputs.scoping(my_scoping)
        """
        return self._scoping

    @property
    def streams_container(self):
        """Allows to connect streams_container input to the operator.

        Needed to get mesh and material ids. optional
        if a data_sources have been
        connected.

        Parameters
        ----------
        my_streams_container : StreamsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.compute_stress_Z()
        >>> op.inputs.streams_container.connect(my_streams_container)
        >>> # or
        >>> op.inputs.streams_container(my_streams_container)
        """
        return self._streams_container

    @property
    def data_sources(self):
        """Allows to connect data_sources input to the operator.

        Needed to get mesh and material ids. optional
        if a streams_container have been
        connected.

        Parameters
        ----------
        my_data_sources : DataSources

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.compute_stress_Z()
        >>> op.inputs.data_sources.connect(my_data_sources)
        >>> # or
        >>> op.inputs.data_sources(my_data_sources)
        """
        return self._data_sources

    @property
    def requested_location(self):
        """Allows to connect requested_location input to the operator.

        Average the elemental nodal result to the
        requested location.

        Parameters
        ----------
        my_requested_location : str

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.compute_stress_Z()
        >>> op.inputs.requested_location.connect(my_requested_location)
        >>> # or
        >>> op.inputs.requested_location(my_requested_location)
        """
        return self._requested_location

    @property
    def strain(self):
        """Allows to connect strain input to the operator.

        Field/or fields container containing only the
        elastic strain field (element nodal).

        Parameters
        ----------
        my_strain : FieldsContainer or Field

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.compute_stress_Z()
        >>> op.inputs.strain.connect(my_strain)
        >>> # or
        >>> op.inputs.strain(my_strain)
        """
        return self._strain


class OutputsComputeStressZ(_Outputs):
    """Intermediate class used to get outputs from
    compute_stress_Z operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.result.compute_stress_Z()
    >>> # Connect inputs : op.inputs. ...
    >>> result_fields_container = op.outputs.fields_container()
    """

    def __init__(self, op: Operator):
        super().__init__(compute_stress_Z._spec().outputs, op)
        self._fields_container = Output(compute_stress_Z._spec().output_pin(0), 0, op)
        self._outputs.append(self._fields_container)

    @property
    def fields_container(self):
        """Allows to get fields_container output of the operator

        Returns
        ----------
        my_fields_container : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.result.compute_stress_Z()
        >>> # Connect inputs : op.inputs. ...
        >>> result_fields_container = op.outputs.fields_container()
        """  # noqa: E501
        return self._fields_container



"""
Animator
========
This module contains the DPF animator class.

Contains classes used to animate results based on workflows using PyVista.
"""
import numpy as np
from typing import Union, Sequence

import ansys.dpf.core as core
from ansys.dpf.core.helpers.utils import _sort_supported_kwargs
from ansys.dpf.core.plotter import _PyVistaPlotter


class _InternalAnimatorFactory:
    """
    Factory for _InternalAnimator based on the backend."""

    @staticmethod
    def get_animator_class():
        return _PyVistaAnimator


class _PyVistaAnimator(_PyVistaPlotter):
    """This _InternalAnimator class is based on PyVista"""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def animate_workflow(
        self,
        loop_over,
        workflow,
        output_name,
        input_name="loop_over",
        save_as="",
        mode_number=None,
        scale_factor=1.0,
        **kwargs,
    ):

        unit = loop_over.unit
        indices = loop_over.scoping.ids

        if scale_factor is None:
            scale_factor = [False] * len(indices)
        type_scale = type(scale_factor)
        if type_scale in [int, float]:
            scale_factor = [scale_factor] * len(indices)
        elif type_scale == list:
            pass
        # elif type_scale in [core.field.Field, core.fields_container.FieldsContainer]:
        #     scale_factor = ["Non-homogenous"]*len(indices)
        else:
            raise ValueError(
                "Argument scale_factor must be an int, a float, or a list of either, "
                f"(not {type_scale})"
            )
        # Initiate movie or gif file if necessary
        if save_as:
            if save_as.endswith(".gif"):
                self._plotter.open_gif(save_as)
            else:  # pragma: no cover
                kwargs_in = _sort_supported_kwargs(bound_method=self._plotter.open_movie, **kwargs)
                try:
                    self._plotter.open_movie(save_as, **kwargs_in)
                except ImportError as e:
                    if "imageio ffmpeg plugin you need" in e.msg:
                        raise ImportError(
                            "The imageio-ffmpeg library is required to save "
                            "animations. Please install it first with the command "
                            "'pip install imageio-ffmpeg'"
                        )
                    else:
                        raise e
        freq_kwargs = kwargs.pop("freq_kwargs", {})
        freq_fmt = freq_kwargs.pop("fmt", "")

        cpos = kwargs.pop("cpos", None)
        if cpos:
            if isinstance(cpos[0][0], float):
                cpos = [cpos] * len(indices)

        def render_frame(frame):
            self._plotter.clear()

            if mode_number is None:
                workflow.connect(input_name, [frame])

            else:
                workflow.connect(input_name, loop_over.data[frame])

            field = workflow.get_output(output_name, core.types.field)
            deform = None
            if "deform_by" in workflow.output_names:
                deform = workflow.get_output("deform_by", core.types.field)
            self.add_field(
                field,
                deform_by=deform,
                scale_factor_legend=scale_factor[frame],
                **kwargs,
            )
            kwargs_in = _sort_supported_kwargs(bound_method=self._plotter.add_text, **freq_kwargs)
            if mode_number is None:
                str_template = "t={0:{2}} {1}"
                self._plotter.add_text(
                    str_template.format(indices[frame], unit, freq_fmt), **kwargs_in
                )
            else:
                str_template = "frq={0:{2}} {1}"
                self._plotter.add_text(
                    str_template.format(mode_number, unit, freq_fmt), **kwargs_in
                )

            if cpos:
                self._plotter.camera_position = cpos[frame]

        try:

            def animation():
                if save_as:
                    try:
                        self._plotter.write_frame()
                    except AttributeError as e:  # pragma: no cover
                        if (
                            "To retrieve an image after the render window has been closed"
                            in e.args[0]
                        ):
                            print("Animation canceled.")
                            print(e)
                            return result
                # For each additional frame requested
                if len(indices) > 1:
                    for frame in range(1, len(indices)):
                        try:
                            render_frame(frame)
                        except AttributeError as e:  # pragma: no cover
                            if "'NoneType' object has no attribute 'interactor'" in e.args[0]:
                                print("Animation canceled.")
                                return result
                        if save_as:
                            self._plotter.write_frame()

            # Write initial frame
            render_frame(0)
            # If not off_screen, enable the user to choose the camera position
            off_screen = kwargs.pop("off_screen", None)
            if off_screen is None:
                import pyvista as pv

                off_screen = pv.OFF_SCREEN

            if not off_screen:
                self._plotter.add_key_event("a", animation)
                print('Orient the view, then press "a" to produce an animation')
            else:
                animation()
            # Show is necessary even when off_screen to initiate the renderer
            result = self._plotter.show(interactive=True)
            # result = self.show_figure(auto_close=False, **kwargs)
            # result = self._plotter.show()
        except Exception as e:  # pragma: no cover
            print(e)
            raise
        self._plotter.close()
        return result


class Animator:
    def __init__(self, workflow=None, **kwargs):
        """
        Create an Animator object.

        The current Animator is a PyVista based object.

        That means that PyVista must be installed, and that
        it supports **kwargs as parameter (the argument
        must be supported by the installed PyVista version).
        More information about the available arguments are
        available at :class:`pyvista.Plotter`.

        Parameters
        ----------
        workflow : Workflow, optional
            Workflow used to generate a Field at each frame of the animation.
            By default, the "to_render" Field output will be plotted,
            and the "loop_over" input defines what the animation iterates on.
            Optionally, the workflow can also have a "deform_by" Field output,
            used to deform the mesh support.
        **kwargs : optional
            Additional keyword arguments for the plotter. More information
            are available at :class:`pyvista.Plotter`.

        Examples
        --------
        >>> from ansys.dpf.core.animator import Animator
        >>> anim = Animator(notebook=False)

        """
        _InternalAnimatorClass = _InternalAnimatorFactory.get_animator_class()
        self._internal_animator = _InternalAnimatorClass(**kwargs)
        self._workflow = workflow

    @property
    def workflow(self) -> core.Workflow:
        """
        Workflow used to generate a Field at each frame of the animation.
        By default, the "to_render" Field output will be plotted,
        and the "loop_over" input defines what the animation iterates on.
        Optionally, the workflow can also have a "deform_by" Field output,
        used to deform the mesh support.

        Returns
        -------
        workflow : Workflow
        """
        return self._workflow

    @workflow.setter
    def workflow(self, workflow: core.Workflow):
        """
        Set the workflow used to generate a Field at each frame of the animation.

        Parameters
        ----------
        workflow : Workflow
            Workflow used to generate a Field at each frame of the animation.
            By default, the "to_render" Field output will be plotted,
            and the "loop_over" input defines what the animation iterates on.
            Optionally, the workflow can also have a "deform_by" Field output,
            used to deform the mesh support.

        """
        self._workflow = workflow

    def animate(
        self,
        loop_over: core.Field,
        output_name: str = "to_render",
        input_name: str = "loop_over",
        save_as: str = None,
        scale_factor: Union[float, Sequence[float]] = 1.0,
        freq_kwargs: dict = None,
        **kwargs,
    ):
        """
        Animate the workflow of the Animator, using inputs

        Parameters
        ----------
        loop_over : Field
            Field of values to loop over.
            Can for example be a subset of sets of TimeFreqSupport.time_frequencies.
            The unit of the Field will be displayed if present.
        output_name : str, optional
            Name of the workflow output to use as Field for each frame's contour.
            Defaults to "to_render".
        input_name : list of str, optional
            Name of the workflow inputs to feed loop_over values into.
            Defaults to "loop_over".
        save_as : str, optional
            Path of file to save the animation to. Defaults to None. Can be of any format supported
            by pyvista.Plotter.write_frame (.gif, .mp4, ...).
        scale_factor : float, list, optional
            Scale factor to apply when warping the mesh. Defaults to 1.0. Can be a list to make
            scaling frequency-dependent.
        freq_kwargs : dict, optional
            Dictionary of kwargs given to the :func:`pyvista.Plotter.add_text` method, used to
            format the frequency information. Can also contain a "fmt" key,
            defining the format for the frequency displayed with a string such as ".3e".
        **kwargs : optional
            Additional keyword arguments for the animator.
            Used by :func:`pyvista.Plotter` (off_screen, cpos, ...),
            or by :func:`pyvista.Plotter.open_movie`
            (framerate, quality, ...)


        """
        if freq_kwargs is None:
            freq_kwargs = {"font_size": 12, "fmt": ".3e"}
        if self.workflow is None:
            raise ValueError("Cannot animate without self.workflow.")
        return self._internal_animator.animate_workflow(
            loop_over=loop_over,
            workflow=self.workflow,
            output_name=output_name,
            input_name=input_name,
            save_as=save_as,
            scale_factor=scale_factor,
            freq_kwargs=freq_kwargs,
            **kwargs,
        )


def scale_factor_to_fc(scale_factor, fc):
    def int_to_field(value, shape, scoping):
        field = core.fields_factory.field_from_array(np.full(shape=shape, fill_value=value))
        field.scoping = scoping
        return field

    scale_type = type(scale_factor)
    n_sets = len(fc)
    if scale_type == core.field.Field:
        raise NotImplementedError("Scaling by a Field is not yet implemented.")
        # # Turn the Field into a fields_container
        # fields = []
        # for i in range(n_sets):
        #     fields.append(scale_factor)
        # scale_factor = core.fields_container_factory.over_time_freq_fields_container(fields)
    elif scale_type == core.fields_container.FieldsContainer:
        raise NotImplementedError("Scaling by a FieldsContainer is not yet implemented.")
        # if scale_factor.time_freq_support.n_sets != n_sets:
        #     raise ValueError(f"The scale_factor FieldsContainer does not contain the same "
        #                      f"number of fields as the fields_container being animated "
        #                      f" ({scale_factor.time_freq_support.n_sets} != {n_sets}).")
    elif scale_type == list:
        if len(scale_factor) != n_sets:
            raise ValueError(
                f"The scale_factor list is not the same length as the fields_container"
                f"being animated ({len(scale_factor)} != {n_sets})."
            )
        # Turn the scalar list into a FieldsContainer
        fields = []
        for i in range(len(fc)):
            fields.append(
                int_to_field(scale_factor[i], fc.get_field(0).shape, fc.get_field(0).scoping)
            )
        scale_factor = core.fields_container_factory.over_time_freq_fields_container(fields)
    elif scale_type == int or scale_type == float:
        # Turn the float into a fields_container
        fields = []
        for i in range(n_sets):
            fields.append(
                int_to_field(scale_factor, fc.get_field(0).shape, fc.get_field(0).scoping)
            )
        scale_factor = core.fields_container_factory.over_time_freq_fields_container(fields)
    else:
        raise ValueError(
            "Argument scale_factor must be an int, a float, or a list of either, "
            f"(not {scale_type})"
        )
    return scale_factor



"""
reduce_sampling
===============
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class reduce_sampling(Operator):
    """Take a scoping and remove half of it's content.

    Parameters
    ----------
    mesh_scoping : Scoping
    denominator : int, optional
        Set the number of time the scoping is reduced
        (default is 2). must be integer value
        above 1.


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.scoping.reduce_sampling()

    >>> # Make input connections
    >>> my_mesh_scoping = dpf.Scoping()
    >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
    >>> my_denominator = int()
    >>> op.inputs.denominator.connect(my_denominator)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.scoping.reduce_sampling(
    ...     mesh_scoping=my_mesh_scoping,
    ...     denominator=my_denominator,
    ... )

    >>> # Get output data
    >>> result_mesh_scoping = op.outputs.mesh_scoping()
    """

    def __init__(self, mesh_scoping=None, denominator=None, config=None, server=None):
        super().__init__(name="scoping::reduce_sampling", config=config, server=server)
        self._inputs = InputsReduceSampling(self)
        self._outputs = OutputsReduceSampling(self)
        if mesh_scoping is not None:
            self.inputs.mesh_scoping.connect(mesh_scoping)
        if denominator is not None:
            self.inputs.denominator.connect(denominator)

    @staticmethod
    def _spec():
        description = """Take a scoping and remove half of it's content."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="mesh_scoping",
                    type_names=["scoping"],
                    optional=False,
                    document="""""",
                ),
                1: PinSpecification(
                    name="denominator",
                    type_names=["int32"],
                    optional=True,
                    document="""Set the number of time the scoping is reduced
        (default is 2). must be integer value
        above 1.""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="mesh_scoping",
                    type_names=["scoping"],
                    optional=False,
                    document="""""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(name="scoping::reduce_sampling", server=server)

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsReduceSampling
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsReduceSampling
        """
        return super().outputs


class InputsReduceSampling(_Inputs):
    """Intermediate class used to connect user inputs to
    reduce_sampling operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.scoping.reduce_sampling()
    >>> my_mesh_scoping = dpf.Scoping()
    >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
    >>> my_denominator = int()
    >>> op.inputs.denominator.connect(my_denominator)
    """

    def __init__(self, op: Operator):
        super().__init__(reduce_sampling._spec().inputs, op)
        self._mesh_scoping = Input(reduce_sampling._spec().input_pin(0), 0, op, -1)
        self._inputs.append(self._mesh_scoping)
        self._denominator = Input(reduce_sampling._spec().input_pin(1), 1, op, -1)
        self._inputs.append(self._denominator)

    @property
    def mesh_scoping(self):
        """Allows to connect mesh_scoping input to the operator.

        Parameters
        ----------
        my_mesh_scoping : Scoping

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.scoping.reduce_sampling()
        >>> op.inputs.mesh_scoping.connect(my_mesh_scoping)
        >>> # or
        >>> op.inputs.mesh_scoping(my_mesh_scoping)
        """
        return self._mesh_scoping

    @property
    def denominator(self):
        """Allows to connect denominator input to the operator.

        Set the number of time the scoping is reduced
        (default is 2). must be integer value
        above 1.

        Parameters
        ----------
        my_denominator : int

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.scoping.reduce_sampling()
        >>> op.inputs.denominator.connect(my_denominator)
        >>> # or
        >>> op.inputs.denominator(my_denominator)
        """
        return self._denominator


class OutputsReduceSampling(_Outputs):
    """Intermediate class used to get outputs from
    reduce_sampling operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.scoping.reduce_sampling()
    >>> # Connect inputs : op.inputs. ...
    >>> result_mesh_scoping = op.outputs.mesh_scoping()
    """

    def __init__(self, op: Operator):
        super().__init__(reduce_sampling._spec().outputs, op)
        self._mesh_scoping = Output(reduce_sampling._spec().output_pin(0), 0, op)
        self._outputs.append(self._mesh_scoping)

    @property
    def mesh_scoping(self):
        """Allows to get mesh_scoping output of the operator

        Returns
        ----------
        my_mesh_scoping : Scoping

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.scoping.reduce_sampling()
        >>> # Connect inputs : op.inputs. ...
        >>> result_mesh_scoping = op.outputs.mesh_scoping()
        """  # noqa: E501
        return self._mesh_scoping



"""
cyclic_support_provider
=======================
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class cyclic_support_provider(Operator):
    """Read the cyclic support (DPF entity containing necessary information
    for expansions) and expands the mesh.

    Parameters
    ----------
    streams_container : StreamsContainer, optional
        Streams containing the result file.
    data_sources : DataSources
        Data sources containing the result file.
    sector_meshed_region : MeshedRegion or MeshesContainer, optional
        Mesh of the first sector.
    expanded_meshed_region : MeshedRegion or MeshesContainer, optional
        If this pin is set, expanding the mesh is not
        necessary.
    sectors_to_expand : Scoping or ScopingsContainer, optional
        Sectors to expand (start at 0), for
        multistage: use scopings container
        with 'stage' label.


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.metadata.cyclic_support_provider()

    >>> # Make input connections
    >>> my_streams_container = dpf.StreamsContainer()
    >>> op.inputs.streams_container.connect(my_streams_container)
    >>> my_data_sources = dpf.DataSources()
    >>> op.inputs.data_sources.connect(my_data_sources)
    >>> my_sector_meshed_region = dpf.MeshedRegion()
    >>> op.inputs.sector_meshed_region.connect(my_sector_meshed_region)
    >>> my_expanded_meshed_region = dpf.MeshedRegion()
    >>> op.inputs.expanded_meshed_region.connect(my_expanded_meshed_region)
    >>> my_sectors_to_expand = dpf.Scoping()
    >>> op.inputs.sectors_to_expand.connect(my_sectors_to_expand)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.metadata.cyclic_support_provider(
    ...     streams_container=my_streams_container,
    ...     data_sources=my_data_sources,
    ...     sector_meshed_region=my_sector_meshed_region,
    ...     expanded_meshed_region=my_expanded_meshed_region,
    ...     sectors_to_expand=my_sectors_to_expand,
    ... )

    >>> # Get output data
    >>> result_cyclic_support = op.outputs.cyclic_support()
    >>> result_sector_meshes = op.outputs.sector_meshes()
    """

    def __init__(
        self,
        streams_container=None,
        data_sources=None,
        sector_meshed_region=None,
        expanded_meshed_region=None,
        sectors_to_expand=None,
        config=None,
        server=None,
    ):
        super().__init__(
            name="mapdl::rst::support_provider_cyclic", config=config, server=server
        )
        self._inputs = InputsCyclicSupportProvider(self)
        self._outputs = OutputsCyclicSupportProvider(self)
        if streams_container is not None:
            self.inputs.streams_container.connect(streams_container)
        if data_sources is not None:
            self.inputs.data_sources.connect(data_sources)
        if sector_meshed_region is not None:
            self.inputs.sector_meshed_region.connect(sector_meshed_region)
        if expanded_meshed_region is not None:
            self.inputs.expanded_meshed_region.connect(expanded_meshed_region)
        if sectors_to_expand is not None:
            self.inputs.sectors_to_expand.connect(sectors_to_expand)

    @staticmethod
    def _spec():
        description = """Read the cyclic support (DPF entity containing necessary information
            for expansions) and expands the mesh."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                3: PinSpecification(
                    name="streams_container",
                    type_names=["streams_container"],
                    optional=True,
                    document="""Streams containing the result file.""",
                ),
                4: PinSpecification(
                    name="data_sources",
                    type_names=["data_sources"],
                    optional=False,
                    document="""Data sources containing the result file.""",
                ),
                7: PinSpecification(
                    name="sector_meshed_region",
                    type_names=["abstract_meshed_region", "meshes_container"],
                    optional=True,
                    document="""Mesh of the first sector.""",
                ),
                15: PinSpecification(
                    name="expanded_meshed_region",
                    type_names=["abstract_meshed_region", "meshes_container"],
                    optional=True,
                    document="""If this pin is set, expanding the mesh is not
        necessary.""",
                ),
                18: PinSpecification(
                    name="sectors_to_expand",
                    type_names=["scoping", "scopings_container", "vector<int32>"],
                    optional=True,
                    document="""Sectors to expand (start at 0), for
        multistage: use scopings container
        with 'stage' label.""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="cyclic_support",
                    type_names=["cyclic_support"],
                    optional=False,
                    document="""""",
                ),
                1: PinSpecification(
                    name="sector_meshes",
                    type_names=["meshes_container"],
                    optional=False,
                    document="""""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(
            name="mapdl::rst::support_provider_cyclic", server=server
        )

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsCyclicSupportProvider
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsCyclicSupportProvider
        """
        return super().outputs


class InputsCyclicSupportProvider(_Inputs):
    """Intermediate class used to connect user inputs to
    cyclic_support_provider operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.metadata.cyclic_support_provider()
    >>> my_streams_container = dpf.StreamsContainer()
    >>> op.inputs.streams_container.connect(my_streams_container)
    >>> my_data_sources = dpf.DataSources()
    >>> op.inputs.data_sources.connect(my_data_sources)
    >>> my_sector_meshed_region = dpf.MeshedRegion()
    >>> op.inputs.sector_meshed_region.connect(my_sector_meshed_region)
    >>> my_expanded_meshed_region = dpf.MeshedRegion()
    >>> op.inputs.expanded_meshed_region.connect(my_expanded_meshed_region)
    >>> my_sectors_to_expand = dpf.Scoping()
    >>> op.inputs.sectors_to_expand.connect(my_sectors_to_expand)
    """

    def __init__(self, op: Operator):
        super().__init__(cyclic_support_provider._spec().inputs, op)
        self._streams_container = Input(
            cyclic_support_provider._spec().input_pin(3), 3, op, -1
        )
        self._inputs.append(self._streams_container)
        self._data_sources = Input(
            cyclic_support_provider._spec().input_pin(4), 4, op, -1
        )
        self._inputs.append(self._data_sources)
        self._sector_meshed_region = Input(
            cyclic_support_provider._spec().input_pin(7), 7, op, -1
        )
        self._inputs.append(self._sector_meshed_region)
        self._expanded_meshed_region = Input(
            cyclic_support_provider._spec().input_pin(15), 15, op, -1
        )
        self._inputs.append(self._expanded_meshed_region)
        self._sectors_to_expand = Input(
            cyclic_support_provider._spec().input_pin(18), 18, op, -1
        )
        self._inputs.append(self._sectors_to_expand)

    @property
    def streams_container(self):
        """Allows to connect streams_container input to the operator.

        Streams containing the result file.

        Parameters
        ----------
        my_streams_container : StreamsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.metadata.cyclic_support_provider()
        >>> op.inputs.streams_container.connect(my_streams_container)
        >>> # or
        >>> op.inputs.streams_container(my_streams_container)
        """
        return self._streams_container

    @property
    def data_sources(self):
        """Allows to connect data_sources input to the operator.

        Data sources containing the result file.

        Parameters
        ----------
        my_data_sources : DataSources

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.metadata.cyclic_support_provider()
        >>> op.inputs.data_sources.connect(my_data_sources)
        >>> # or
        >>> op.inputs.data_sources(my_data_sources)
        """
        return self._data_sources

    @property
    def sector_meshed_region(self):
        """Allows to connect sector_meshed_region input to the operator.

        Mesh of the first sector.

        Parameters
        ----------
        my_sector_meshed_region : MeshedRegion or MeshesContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.metadata.cyclic_support_provider()
        >>> op.inputs.sector_meshed_region.connect(my_sector_meshed_region)
        >>> # or
        >>> op.inputs.sector_meshed_region(my_sector_meshed_region)
        """
        return self._sector_meshed_region

    @property
    def expanded_meshed_region(self):
        """Allows to connect expanded_meshed_region input to the operator.

        If this pin is set, expanding the mesh is not
        necessary.

        Parameters
        ----------
        my_expanded_meshed_region : MeshedRegion or MeshesContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.metadata.cyclic_support_provider()
        >>> op.inputs.expanded_meshed_region.connect(my_expanded_meshed_region)
        >>> # or
        >>> op.inputs.expanded_meshed_region(my_expanded_meshed_region)
        """
        return self._expanded_meshed_region

    @property
    def sectors_to_expand(self):
        """Allows to connect sectors_to_expand input to the operator.

        Sectors to expand (start at 0), for
        multistage: use scopings container
        with 'stage' label.

        Parameters
        ----------
        my_sectors_to_expand : Scoping or ScopingsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.metadata.cyclic_support_provider()
        >>> op.inputs.sectors_to_expand.connect(my_sectors_to_expand)
        >>> # or
        >>> op.inputs.sectors_to_expand(my_sectors_to_expand)
        """
        return self._sectors_to_expand


class OutputsCyclicSupportProvider(_Outputs):
    """Intermediate class used to get outputs from
    cyclic_support_provider operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.metadata.cyclic_support_provider()
    >>> # Connect inputs : op.inputs. ...
    >>> result_cyclic_support = op.outputs.cyclic_support()
    >>> result_sector_meshes = op.outputs.sector_meshes()
    """

    def __init__(self, op: Operator):
        super().__init__(cyclic_support_provider._spec().outputs, op)
        self._cyclic_support = Output(
            cyclic_support_provider._spec().output_pin(0), 0, op
        )
        self._outputs.append(self._cyclic_support)
        self._sector_meshes = Output(
            cyclic_support_provider._spec().output_pin(1), 1, op
        )
        self._outputs.append(self._sector_meshes)

    @property
    def cyclic_support(self):
        """Allows to get cyclic_support output of the operator

        Returns
        ----------
        my_cyclic_support : CyclicSupport

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.metadata.cyclic_support_provider()
        >>> # Connect inputs : op.inputs. ...
        >>> result_cyclic_support = op.outputs.cyclic_support()
        """  # noqa: E501
        return self._cyclic_support

    @property
    def sector_meshes(self):
        """Allows to get sector_meshes output of the operator

        Returns
        ----------
        my_sector_meshes : MeshesContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.metadata.cyclic_support_provider()
        >>> # Connect inputs : op.inputs. ...
        >>> result_sector_meshes = op.outputs.sector_meshes()
        """  # noqa: E501
        return self._sector_meshes



"""
fc_get_attribute
================
Autogenerated DPF operator classes.
"""
from warnings import warn
from ansys.dpf.core.dpf_operator import Operator
from ansys.dpf.core.inputs import Input, _Inputs
from ansys.dpf.core.outputs import Output, _Outputs
from ansys.dpf.core.outputs import _modify_output_spec_with_one_type
from ansys.dpf.core.operators.specification import PinSpecification, Specification


class fc_get_attribute(Operator):
    """Uses the FieldsContainer APIs to return a given attribute of the mesh
    in input.

    Parameters
    ----------
    fields_container : FieldsContainer
    property_name : str
        Supported property names are:
        "label_scoping", "label_values",
        "time_freq_support", "labels",
        "field_scoping" .
    property_identifier : str or int, optional
        Additional pin for some property : the label
        name for "label_scoping" or
        "label_values", the field index
        (default 0) for "field_scoping".


    Examples
    --------
    >>> from ansys.dpf import core as dpf

    >>> # Instantiate operator
    >>> op = dpf.operators.utility.fc_get_attribute()

    >>> # Make input connections
    >>> my_fields_container = dpf.FieldsContainer()
    >>> op.inputs.fields_container.connect(my_fields_container)
    >>> my_property_name = str()
    >>> op.inputs.property_name.connect(my_property_name)
    >>> my_property_identifier = str()
    >>> op.inputs.property_identifier.connect(my_property_identifier)

    >>> # Instantiate operator and connect inputs in one line
    >>> op = dpf.operators.utility.fc_get_attribute(
    ...     fields_container=my_fields_container,
    ...     property_name=my_property_name,
    ...     property_identifier=my_property_identifier,
    ... )

    >>> # Get output data
    >>> result_property = op.outputs.property()
    """

    def __init__(
        self,
        fields_container=None,
        property_name=None,
        property_identifier=None,
        config=None,
        server=None,
    ):
        super().__init__(
            name="fieldscontainer::get_attribute", config=config, server=server
        )
        self._inputs = InputsFcGetAttribute(self)
        self._outputs = OutputsFcGetAttribute(self)
        if fields_container is not None:
            self.inputs.fields_container.connect(fields_container)
        if property_name is not None:
            self.inputs.property_name.connect(property_name)
        if property_identifier is not None:
            self.inputs.property_identifier.connect(property_identifier)

    @staticmethod
    def _spec():
        description = """Uses the FieldsContainer APIs to return a given attribute of the mesh
            in input."""
        spec = Specification(
            description=description,
            map_input_pin_spec={
                0: PinSpecification(
                    name="fields_container",
                    type_names=["fields_container"],
                    optional=False,
                    document="""""",
                ),
                1: PinSpecification(
                    name="property_name",
                    type_names=["string"],
                    optional=False,
                    document="""Supported property names are:
        "label_scoping", "label_values",
        "time_freq_support", "labels",
        "field_scoping" .""",
                ),
                2: PinSpecification(
                    name="property_identifier",
                    type_names=["string", "int32"],
                    optional=True,
                    document="""Additional pin for some property : the label
        name for "label_scoping" or
        "label_values", the field index
        (default 0) for "field_scoping".""",
                ),
            },
            map_output_pin_spec={
                0: PinSpecification(
                    name="property",
                    type_names=[
                        "scoping",
                        "vector<int32>",
                        "time_freq_support",
                        "vector<string>",
                        "label_space",
                    ],
                    optional=False,
                    document="""Returns a scoping for property:
        "label_scoping", a vector of int for
        "label_values", a time freq support
        for "time_freq_support", a vector of
        string for "labels", a labelspace for
        "field_scoping".""",
                ),
            },
        )
        return spec

    @staticmethod
    def default_config(server=None):
        """Returns the default config of the operator.

        This config can then be changed to the user needs and be used to
        instantiate the operator. The Configuration allows to customize
        how the operation will be processed by the operator.

        Parameters
        ----------
        server : server.DPFServer, optional
            Server with channel connected to the remote or local instance. When
            ``None``, attempts to use the global server.
        """
        return Operator.default_config(
            name="fieldscontainer::get_attribute", server=server
        )

    @property
    def inputs(self):
        """Enables to connect inputs to the operator

        Returns
        --------
        inputs : InputsFcGetAttribute
        """
        return super().inputs

    @property
    def outputs(self):
        """Enables to get outputs of the operator by evaluating it

        Returns
        --------
        outputs : OutputsFcGetAttribute
        """
        return super().outputs


class InputsFcGetAttribute(_Inputs):
    """Intermediate class used to connect user inputs to
    fc_get_attribute operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.utility.fc_get_attribute()
    >>> my_fields_container = dpf.FieldsContainer()
    >>> op.inputs.fields_container.connect(my_fields_container)
    >>> my_property_name = str()
    >>> op.inputs.property_name.connect(my_property_name)
    >>> my_property_identifier = str()
    >>> op.inputs.property_identifier.connect(my_property_identifier)
    """

    def __init__(self, op: Operator):
        super().__init__(fc_get_attribute._spec().inputs, op)
        self._fields_container = Input(fc_get_attribute._spec().input_pin(0), 0, op, -1)
        self._inputs.append(self._fields_container)
        self._property_name = Input(fc_get_attribute._spec().input_pin(1), 1, op, -1)
        self._inputs.append(self._property_name)
        self._property_identifier = Input(
            fc_get_attribute._spec().input_pin(2), 2, op, -1
        )
        self._inputs.append(self._property_identifier)

    @property
    def fields_container(self):
        """Allows to connect fields_container input to the operator.

        Parameters
        ----------
        my_fields_container : FieldsContainer

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.fc_get_attribute()
        >>> op.inputs.fields_container.connect(my_fields_container)
        >>> # or
        >>> op.inputs.fields_container(my_fields_container)
        """
        return self._fields_container

    @property
    def property_name(self):
        """Allows to connect property_name input to the operator.

        Supported property names are:
        "label_scoping", "label_values",
        "time_freq_support", "labels",
        "field_scoping" .

        Parameters
        ----------
        my_property_name : str

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.fc_get_attribute()
        >>> op.inputs.property_name.connect(my_property_name)
        >>> # or
        >>> op.inputs.property_name(my_property_name)
        """
        return self._property_name

    @property
    def property_identifier(self):
        """Allows to connect property_identifier input to the operator.

        Additional pin for some property : the label
        name for "label_scoping" or
        "label_values", the field index
        (default 0) for "field_scoping".

        Parameters
        ----------
        my_property_identifier : str or int

        Examples
        --------
        >>> from ansys.dpf import core as dpf
        >>> op = dpf.operators.utility.fc_get_attribute()
        >>> op.inputs.property_identifier.connect(my_property_identifier)
        >>> # or
        >>> op.inputs.property_identifier(my_property_identifier)
        """
        return self._property_identifier


class OutputsFcGetAttribute(_Outputs):
    """Intermediate class used to get outputs from
    fc_get_attribute operator.

    Examples
    --------
    >>> from ansys.dpf import core as dpf
    >>> op = dpf.operators.utility.fc_get_attribute()
    >>> # Connect inputs : op.inputs. ...
    >>> result_property = op.outputs.property()
    """

    def __init__(self, op: Operator):
        super().__init__(fc_get_attribute._spec().outputs, op)
        self.property_as_scoping = Output(
            _modify_output_spec_with_one_type(
                fc_get_attribute._spec().output_pin(0), "scoping"
            ),
            0,
            op,
        )
        self._outputs.append(self.property_as_scoping)
        self.property_as_vector_int32_ = Output(
            _modify_output_spec_with_one_type(
                fc_get_attribute._spec().output_pin(0), "vector_int32_"
            ),
            0,
            op,
        )
        self._outputs.append(self.property_as_vector_int32_)
        self.property_as_time_freq_support = Output(
            _modify_output_spec_with_one_type(
                fc_get_attribute._spec().output_pin(0), "time_freq_support"
            ),
            0,
            op,
        )
        self._outputs.append(self.property_as_time_freq_support)
        self.property_as_vector_string_ = Output(
            _modify_output_spec_with_one_type(
                fc_get_attribute._spec().output_pin(0), "vector_string_"
            ),
            0,
            op,
        )
        self._outputs.append(self.property_as_vector_string_)
        self.property_as_label_space = Output(
            _modify_output_spec_with_one_type(
                fc_get_attribute._spec().output_pin(0), "label_space"
            ),
            0,
            op,
        )
        self._outputs.append(self.property_as_label_space)
